from attrs import frozen, Factory


@frozen
class Disconnected:
    pass


@frozen
class Connecting:
    url: str
    last_attempt: int  # unix-timestamp when we last tried connecting


@frozen
class Connected:
    url: str


@frozen
class Failed:
    reason: str


@frozen
class Closed:
    """
    We have purposely shut down our connection to the server
    """


@frozen
class NoPeer:
    """
    We have yet to see a peer
    """


@frozen
class StoppedPeer:
    """
    We have disconnected from the peer (probably on purpose)
    """


@frozen
class ConnectingPeer:
    """
    We are actively trying to connect to a peer
    """
    last_attempt: int  # unix-timestamp


@frozen
class ReconnectingPeer:
    """
    We are actively trying to connect to a peer.
    In contract to `ConnectingPeer`, we've already reached our peer at least once.
    """
    last_attempt: int  # unix-timestamp


@frozen
class ConnectedPeer:
    connected_at: int  # when we first connected (seconds since epoch)
    expires_at: int  # earliest we will consider re-connecting (seconds since epoch)
    hint_description: str  # what sort of connection is this?


@frozen
class NoKey:
    pass


@frozen
class AllegedSharedKey:
    pass  # DO NOT reveal alleged key here; "status" messages are for users


@frozen
class ConfirmedKey:
    pass  # DO NOT reveal real key here; "status" messages are for users


@frozen
class NoCode:
    """
    Not allocated yet
    """


@frozen
class AllocatedCode:
    """
    A valid code is available
    """


@frozen
class ConsumedCode:
    """
    The code was used by the other side, and is now no longer
    usable. The nameplate has been un-claimed (and may be reused by a
    different code any time).
    """


# note that the tests (e.g. test_api.py) leverage the order of these
# types inside the Unions, so that we don't have to separately
# enumerate the expected order (and then also ensure we don't miss any)


# General mailbox statuses
ConnectionStatus = Disconnected | Connecting | Connected | Failed | Closed
PeerSharedKey = NoKey | AllegedSharedKey | ConfirmedKey
CodeStatus = NoCode | AllocatedCode | ConsumedCode

# Dilation only
PeerConnection = NoPeer | ConnectingPeer | ConnectedPeer | ReconnectingPeer | StoppedPeer


# NOTE: probably none of the status stuff should ever reveal secret or
# sensitive information -- on the grounds all this will probably be
# shown to a user at some point


@frozen
class WormholeStatus:
    """
    Represents the current status of a wormhole for use by the outside
    """

    # are we connected to the Mailbox Server?
    mailbox_connection: ConnectionStatus = Disconnected()

    # only Dilation (or "transit") know if we've actually achieved a
    # connection to our peer; this just tracks the PAKE negotiation,
    # basically
    peer_key: PeerSharedKey = NoKey()

    # we don't reveal the actual code here, on the theory the UI
    # should already know it and/or be displaying it somehow. This
    # communicates the *status* of that code, e.g. whether it is stale
    # or not.
    code: CodeStatus = NoCode()


@frozen
class DilationHint:
    url: str
    is_direct: bool


@frozen
class DilationStatus:
    """
    Represents the current status of a Dilated wormhole

    """
    # If we are using Dilation (or trying to) it definitely means we
    # have a Wormhole, and thus a WormholeStatus too -- but do we
    # actually want to 'embed' the wormhole status like this?

    # status of our connection to the Mailbox Server
    mailbox: WormholeStatus

    # current Dilation generation (ever increasing number)
    generation: int

    # communication status with peer
    peer_connection: PeerConnection = NoPeer()

    # available methods to get to peer
    hints: set[DilationHint] = Factory(set)



# inspired-by/adapted-from Foolscap's eventual.py, which Glyph wrote for me
# years ago.

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IReactorTime
from twisted.python import log


class EventualQueue:
    def __init__(self, clock):
        # pass clock=reactor unless you're testing
        self._clock = IReactorTime(clock)
        self._calls = []
        self._flush_d = None
        self._timer = None

    def eventually(self, f, *args, **kwargs):
        self._calls.append((f, args, kwargs))
        if not self._timer:
            self._timer = self._clock.callLater(0, self._turn)

    def fire_eventually(self, value=None):
        d = Deferred()
        self.eventually(d.callback, value)
        return d

    def _turn(self):
        self._calls, to_call = [], self._calls
        for f, args, kwargs in to_call:
            try:
                f(*args, **kwargs)
            except Exception:
                log.err()
        self._timer = None

        # Since the only guidance about semantics is the comment about
        # Foolscap, we make sure that any calls added by the above
        # (i.e. by the calls we just ran) only run in the _next_ turn
        # (as Foolscap does). Not doing this leads to some unexpected
        # dependency of the tests on the precise order things are run
        # in a single turn, which defeats the purpose of this
        # "eventual queue".
        if len(self._calls):
            self._timer = self._clock.callLater(0, self._turn)
        else:
            d, self._flush_d = self._flush_d, None
            if d:
                d.callback(None)

    def flush_sync(self):
        # if you have control over the Clock, this will synchronously flush the
        # queue
        assert self._clock.advance, "needs clock=twisted.internet.task.Clock()"
        while self._calls:
            self._clock.advance(0)

    def flush(self):
        # this is for unit tests, not application code
        assert not self._flush_d, "only one flush at a time"
        self._flush_d = Deferred()
        self.eventually(lambda: None)
        return self._flush_d



from twisted.internet.task import Clock
from twisted.python.failure import Failure
import pytest
import pytest_twisted

from ..eventual import EventualQueue
from ..observer import OneShotObserver, SequenceObserver, EmptyableSet


@pytest_twisted.ensureDeferred()
async def test_fire():
    c = Clock()
    eq = EventualQueue(c)
    o = OneShotObserver(eq)
    res = object()
    d1 = o.when_fired()
    eq.flush_sync()
    assert not d1.called
    o.fire(res)
    eq.flush_sync()
    await d1 is res
    d2 = o.when_fired()
    eq.flush_sync()
    await d2 is res
    o.fire_if_not_fired(object())
    eq.flush_sync()

@pytest_twisted.ensureDeferred()
async def test_fire_if_not_fired():
    c = Clock()
    eq = EventualQueue(c)
    o = OneShotObserver(eq)
    res1 = object()
    res2 = object()
    d1 = o.when_fired()
    eq.flush_sync()
    assert not d1.called
    o.fire_if_not_fired(res1)
    o.fire_if_not_fired(res2)
    eq.flush_sync()
    await d1 is res1


@pytest_twisted.ensureDeferred()
async def test_error_before_firing():
    c = Clock()
    eq = EventualQueue(c)
    o = OneShotObserver(eq)
    f = Failure(ValueError("oops"))
    d1 = o.when_fired()
    eq.flush_sync()
    assert not d1.called
    o.error(f)
    eq.flush_sync()
    with pytest.raises(Exception) as ar:
        await d1
    assert ar.value is f.value
    d2 = o.when_fired()
    eq.flush_sync()
    with pytest.raises(Exception) as ar:
        await d2
    assert ar.value is f.value


@pytest_twisted.ensureDeferred()
async def test_error_after_firing():
    c = Clock()
    eq = EventualQueue(c)
    o = OneShotObserver(eq)
    res = object()
    f = Failure(ValueError("oops"))

    o.fire(res)
    eq.flush_sync()
    d1 = o.when_fired()
    o.error(f)
    d2 = o.when_fired()
    eq.flush_sync()
    await d1 is res
    with pytest.raises(Exception) as ar:
        await d2
    assert ar.value is f.value


@pytest_twisted.ensureDeferred()
async def test_fire_multiple():
    c = Clock()
    eq = EventualQueue(c)
    o = SequenceObserver(eq)
    d1 = o.when_next_event()
    eq.flush_sync()
    assert not d1.called
    d2 = o.when_next_event()
    eq.flush_sync()
    assert not d1.called
    assert not d2.called

    ev1 = object()
    ev2 = object()
    o.fire(ev1)
    eq.flush_sync()
    await d1 is ev1
    assert not d2.called

    o.fire(ev2)
    eq.flush_sync()
    await d2 is ev2

    ev3 = object()
    ev4 = object()
    o.fire(ev3)
    o.fire(ev4)

    d3 = o.when_next_event()
    eq.flush_sync()
    await d3 is ev3

    d4 = o.when_next_event()
    eq.flush_sync()
    await d4 is ev4


@pytest_twisted.ensureDeferred()
async def test_error():
    c = Clock()
    eq = EventualQueue(c)
    o = SequenceObserver(eq)
    d1 = o.when_next_event()
    eq.flush_sync()
    assert not d1.called
    f = Failure(ValueError("oops"))
    o.fire(f)
    eq.flush_sync()
    with pytest.raises(Exception) as ar:
        await d1
    assert ar.value is f.value
    d2 = o.when_next_event()
    eq.flush_sync()
    with pytest.raises(Exception) as ar:
        await d2
    assert ar.value is f.value


@pytest_twisted.ensureDeferred()
async def test_set():
    eq = EventualQueue(Clock())
    s = EmptyableSet(_eventual_queue=eq)
    d1 = s.when_next_empty()
    eq.flush_sync()
    assert not d1.called
    s.add(1)
    eq.flush_sync()
    assert not d1.called
    s.add(2)
    s.discard(1)
    d2 = s.when_next_empty()
    eq.flush_sync()
    assert not d1.called
    assert not d2.called
    s.discard(2)
    eq.flush_sync()
    assert await d1 is None
    assert await d2 is None

    s.add(3)
    s.discard(3)



# We use 'threading' defensively here, to detect if we're being called from a
# non-main thread. _rlcompleter.py is the only internal Wormhole code that
# deliberately creates a new thread.
import threading

from attr import attrib, attrs
from automat import MethodicalMachine
from twisted.internet import defer
from zope.interface import implementer

from . import _interfaces, errors
from ._nameplate import validate_nameplate
from .util import provides


def first(outputs):
    return list(outputs)[0]


@attrs
@implementer(_interfaces.IInput)
class Input:
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._all_nameplates = set()
        self._nameplate = None
        self._wordlist = None
        self._wordlist_waiters = []
        self._trace = None

    def set_debug(self, f):
        self._trace = f

    def _debug(self, what):  # pragma: no cover
        if self._trace:
            self._trace(old_state="", input=what, new_state="")

    def wire(self, code, lister):
        self._C = _interfaces.ICode(code)
        self._L = _interfaces.ILister(lister)

    def when_wordlist_is_available(self):
        if self._wordlist:
            return defer.succeed(None)
        d = defer.Deferred()
        self._wordlist_waiters.append(d)
        return d

    @m.state(initial=True)
    def S0_idle(self):
        pass  # pragma: no cover

    @m.state()
    def S1_typing_nameplate(self):
        pass  # pragma: no cover

    @m.state()
    def S2_typing_code_no_wordlist(self):
        pass  # pragma: no cover

    @m.state()
    def S3_typing_code_yes_wordlist(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def S4_done(self):
        pass  # pragma: no cover

    # from Code
    @m.input()
    def start(self):
        pass

    # from Lister
    @m.input()
    def got_nameplates(self, all_nameplates):
        pass

    # from Nameplate
    @m.input()
    def got_wordlist(self, wordlist):
        pass

    # API provided to app as ICodeInputHelper
    @m.input()
    def refresh_nameplates(self):
        pass

    @m.input()
    def get_nameplate_completions(self, prefix):
        pass

    def choose_nameplate(self, nameplate):
        validate_nameplate(nameplate)  # can raise KeyFormatError
        self._choose_nameplate(nameplate)

    @m.input()
    def _choose_nameplate(self, nameplate):
        pass

    @m.input()
    def get_word_completions(self, prefix):
        pass

    @m.input()
    def choose_words(self, words):
        pass

    @m.output()
    def do_start(self):
        self._start_timing = self._timing.add("input code", waiting="user")
        self._L.refresh()
        return Helper(self)

    @m.output()
    def do_refresh(self):
        self._L.refresh()

    @m.output()
    def record_nameplates(self, all_nameplates):
        # we get a set of nameplate id strings
        self._all_nameplates = all_nameplates

    @m.output()
    def _get_nameplate_completions(self, prefix):
        completions = set()
        for nameplate in self._all_nameplates:
            if nameplate.startswith(prefix):
                # TODO: it's a little weird that Input is responsible for the
                # hyphen on nameplates, but WordList owns it for words
                completions.add(nameplate + "-")
        return completions

    @m.output()
    def record_all_nameplates(self, nameplate):
        self._nameplate = nameplate
        self._C.got_nameplate(nameplate)

    @m.output()
    def record_wordlist(self, wordlist):
        from ._rlcompleter import debug
        debug("  -record_wordlist")
        self._wordlist = wordlist

    @m.output()
    def notify_wordlist_waiters(self, wordlist):
        while self._wordlist_waiters:
            d = self._wordlist_waiters.pop()
            d.callback(None)

    @m.output()
    def no_word_completions(self, prefix):
        return set()

    @m.output()
    def _get_word_completions(self, prefix):
        assert self._wordlist
        return self._wordlist.get_completions(prefix)

    @m.output()
    def raise_must_choose_nameplate1(self, prefix):
        raise errors.MustChooseNameplateFirstError()

    @m.output()
    def raise_must_choose_nameplate2(self, words):
        raise errors.MustChooseNameplateFirstError()

    @m.output()
    def raise_already_chose_nameplate1(self):
        raise errors.AlreadyChoseNameplateError()

    @m.output()
    def raise_already_chose_nameplate2(self, prefix):
        raise errors.AlreadyChoseNameplateError()

    @m.output()
    def raise_already_chose_nameplate3(self, nameplate):
        raise errors.AlreadyChoseNameplateError()

    @m.output()
    def raise_already_chose_words1(self, prefix):
        raise errors.AlreadyChoseWordsError()

    @m.output()
    def raise_already_chose_words2(self, words):
        raise errors.AlreadyChoseWordsError()

    @m.output()
    def do_words(self, words):
        code = self._nameplate + "-" + words
        self._start_timing.finish()
        self._C.finished_input(code)

    S0_idle.upon(
        start, enter=S1_typing_nameplate, outputs=[do_start], collector=first)
    # wormholes that don't use input_code (i.e. they use allocate_code or
    # generate_code) will never start() us, but Nameplate will give us a
    # wordlist anyways (as soon as the nameplate is claimed), so handle it.
    S0_idle.upon(
        got_wordlist,
        enter=S0_idle,
        outputs=[record_wordlist, notify_wordlist_waiters])
    S1_typing_nameplate.upon(
        got_nameplates, enter=S1_typing_nameplate, outputs=[record_nameplates])
    # but wormholes that *do* use input_code should not get got_wordlist
    # until after we tell Code that we got_nameplate, which is the earliest
    # it can be claimed
    S1_typing_nameplate.upon(
        refresh_nameplates, enter=S1_typing_nameplate, outputs=[do_refresh])
    S1_typing_nameplate.upon(
        get_nameplate_completions,
        enter=S1_typing_nameplate,
        outputs=[_get_nameplate_completions],
        collector=first)
    S1_typing_nameplate.upon(
        _choose_nameplate,
        enter=S2_typing_code_no_wordlist,
        outputs=[record_all_nameplates])
    S1_typing_nameplate.upon(
        get_word_completions,
        enter=S1_typing_nameplate,
        outputs=[raise_must_choose_nameplate1])
    S1_typing_nameplate.upon(
        choose_words,
        enter=S1_typing_nameplate,
        outputs=[raise_must_choose_nameplate2])

    S2_typing_code_no_wordlist.upon(
        got_nameplates, enter=S2_typing_code_no_wordlist, outputs=[])
    S2_typing_code_no_wordlist.upon(
        got_wordlist,
        enter=S3_typing_code_yes_wordlist,
        outputs=[record_wordlist, notify_wordlist_waiters])
    S2_typing_code_no_wordlist.upon(
        refresh_nameplates,
        enter=S2_typing_code_no_wordlist,
        outputs=[raise_already_chose_nameplate1])
    S2_typing_code_no_wordlist.upon(
        get_nameplate_completions,
        enter=S2_typing_code_no_wordlist,
        outputs=[raise_already_chose_nameplate2])
    S2_typing_code_no_wordlist.upon(
        _choose_nameplate,
        enter=S2_typing_code_no_wordlist,
        outputs=[raise_already_chose_nameplate3])
    S2_typing_code_no_wordlist.upon(
        get_word_completions,
        enter=S2_typing_code_no_wordlist,
        outputs=[no_word_completions],
        collector=first)
    S2_typing_code_no_wordlist.upon(
        choose_words, enter=S4_done, outputs=[do_words])

    S3_typing_code_yes_wordlist.upon(
        got_nameplates, enter=S3_typing_code_yes_wordlist, outputs=[])
    # got_wordlist: should never happen
    S3_typing_code_yes_wordlist.upon(
        refresh_nameplates,
        enter=S3_typing_code_yes_wordlist,
        outputs=[raise_already_chose_nameplate1])
    S3_typing_code_yes_wordlist.upon(
        get_nameplate_completions,
        enter=S3_typing_code_yes_wordlist,
        outputs=[raise_already_chose_nameplate2])
    S3_typing_code_yes_wordlist.upon(
        _choose_nameplate,
        enter=S3_typing_code_yes_wordlist,
        outputs=[raise_already_chose_nameplate3])
    S3_typing_code_yes_wordlist.upon(
        get_word_completions,
        enter=S3_typing_code_yes_wordlist,
        outputs=[_get_word_completions],
        collector=first)
    S3_typing_code_yes_wordlist.upon(
        choose_words, enter=S4_done, outputs=[do_words])

    S4_done.upon(got_nameplates, enter=S4_done, outputs=[])
    S4_done.upon(got_wordlist, enter=S4_done, outputs=[])
    S4_done.upon(
        refresh_nameplates,
        enter=S4_done,
        outputs=[raise_already_chose_nameplate1])
    S4_done.upon(
        get_nameplate_completions,
        enter=S4_done,
        outputs=[raise_already_chose_nameplate2])
    S4_done.upon(
        _choose_nameplate,
        enter=S4_done,
        outputs=[raise_already_chose_nameplate3])
    S4_done.upon(
        get_word_completions,
        enter=S4_done,
        outputs=[raise_already_chose_words1])
    S4_done.upon(
        choose_words, enter=S4_done, outputs=[raise_already_chose_words2])


# we only expose the Helper to application code, not _Input
@attrs
@implementer(_interfaces.IInputHelper)
class Helper:
    _input = attrib()

    def __attrs_post_init__(self):
        self._main_thread = threading.current_thread().ident

    def refresh_nameplates(self):
        assert threading.current_thread().ident == self._main_thread
        self._input.refresh_nameplates()

    def get_nameplate_completions(self, prefix):
        assert threading.current_thread().ident == self._main_thread
        return self._input.get_nameplate_completions(prefix)

    def choose_nameplate(self, nameplate):
        assert threading.current_thread().ident == self._main_thread
        self._input._debug("I.choose_nameplate")
        self._input.choose_nameplate(nameplate)
        self._input._debug("I.choose_nameplate finished")

    def when_wordlist_is_available(self):
        assert threading.current_thread().ident == self._main_thread
        return self._input.when_wordlist_is_available()

    def get_word_completions(self, prefix):
        assert threading.current_thread().ident == self._main_thread
        return self._input.get_word_completions(prefix)

    def choose_words(self, words):
        assert threading.current_thread().ident == self._main_thread
        self._input._debug("I.choose_words")
        self._input.choose_words(words)
        self._input._debug("I.choose_words finished")



from unittest import mock
from ..._dilation.connection import (parse_record, encode_record,
                                     KCM, Ping, Pong, Open, Data, Close, Ack)
import pytest


def test_parse():
    assert parse_record(b"\x00") == KCM()
    assert parse_record(b"\x01\x55\x44\x33\x22") == \
                     Ping(ping_id=b"\x55\x44\x33\x22")
    assert parse_record(b"\x02\x55\x44\x33\x22") == \
                     Pong(ping_id=b"\x55\x44\x33\x22")
    assert parse_record(b"\x03\x00\x00\x02\x01\x00\x00\x01\x00proto") == \
                     Open(scid=513, seqnum=256, subprotocol="proto")
    assert parse_record(b"\x04\x00\x00\x02\x02\x00\x00\x01\x01dataaa") == \
                     Data(scid=514, seqnum=257, data=b"dataaa")
    assert parse_record(b"\x05\x00\x00\x02\x03\x00\x00\x01\x02") == \
                     Close(scid=515, seqnum=258)
    assert parse_record(b"\x06\x00\x00\x01\x03") == \
                     Ack(resp_seqnum=259)
    with mock.patch("wormhole._dilation.connection.log.err") as le:
        with pytest.raises(ValueError):
            parse_record(b"\x07unknown")
    assert le.mock_calls == \
                     [mock.call("received unknown message type: {}".format(
                         b"\x07unknown"))]


def test_encode():
    assert encode_record(KCM()) == b"\x00"
    assert encode_record(Ping(ping_id=b"ping")) == b"\x01ping"
    assert encode_record(Pong(ping_id=b"pong")) == b"\x02pong"
    assert encode_record(Open(scid=65536, seqnum=16, subprotocol="proto")) == \
                     b"\x03\x00\x01\x00\x00\x00\x00\x00\x10proto"
    assert encode_record(Data(scid=65537, seqnum=17, data=b"dataaa")) == \
                     b"\x04\x00\x01\x00\x01\x00\x00\x00\x11dataaa"
    assert encode_record(Close(scid=65538, seqnum=18)) == \
                     b"\x05\x00\x01\x00\x02\x00\x00\x00\x12"
    assert encode_record(Ack(resp_seqnum=19)) == \
                     b"\x06\x00\x00\x00\x13"
    with pytest.raises(TypeError) as ar:
        encode_record("not a record")
    assert str(ar.value) == "not a record"



import pytest
from ..._dilation.manager import TrafficTimer

# Correct operation of the TrafficTiming state-machine

class TimerFixture:
    def __init__(self):
        self.reconnects = []
        self.timers = []

        self.timer_count = 0
        self.reconnect_count = 0
        self.t = TrafficTimer(self.do_reconnect, self.start_timer)

    def start_timer(self):
        self.timers.append(self.timer_count)
        self.timer_count += 1

    def do_reconnect(self):
        self.reconnects.append(self.reconnect_count)
        self.reconnect_count += 1


@pytest.fixture()
def traffic_timer():
    yield TimerFixture()


def test_start(traffic_timer):
    """
    We immedialte start a timer after connecting
    """
    traffic_timer.t.got_connection()
    assert traffic_timer.timers == [0]


def test_one_interval(traffic_timer):
    """
    No reconnect when only a single interval has passed
    """
    traffic_timer.t.got_connection()
    traffic_timer.t.interval_elapsed()
    # timer must be re-started after one elapsed interval
    assert traffic_timer.timers == [0, 1]
    assert traffic_timer.reconnects == []

def test_expired_connection(traffic_timer):
    """
    Trigger a reconnect after two intervals pass with no traffic seen
    """
    traffic_timer.t.got_connection()
    traffic_timer.t.interval_elapsed()
    traffic_timer.t.interval_elapsed()
    # timer _not_ re-started after second interval, because now
    # we're re-connecting
    assert traffic_timer.timers == [0, 1]
    assert traffic_timer.reconnects == [0]

    # at some point after the re-connect trigger we would lose our
    # connection
    traffic_timer.t.lost_connection()
    assert traffic_timer.timers == [0, 1]
    assert traffic_timer.reconnects == [0]



import os
import sys

from attr import attrib, attrs
from twisted.python import failure
from twisted.internet.task import Cooperator
from zope.interface import implementer

from ._boss import Boss
from ._dilation.manager import DILATION_VERSIONS
from ._dilation.connector import Connector
from ._interfaces import IDeferredWormhole, IWormhole
from ._key import derive_key
from .errors import NoKeyError, WormholeClosed
from .eventual import EventualQueue
from .journal import ImmediateJournal
from .observer import OneShotObserver, SequenceObserver
from .timing import DebugTiming
from .util import bytes_to_hexstr, to_bytes
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

# We can provide different APIs to different apps:
# * Deferreds
#   w.get_code().addCallback(print_code)
#   w.send_message(data)
#   w.get_message().addCallback(got_data)
#   w.close().addCallback(closed)

# * delegate callbacks (better for journaled environments)
#   w = wormhole(delegate=app)
#   w.send_message(data)
#   app.wormhole_got_code(code)
#   app.wormhole_got_verifier(verifier)
#   app.wormhole_got_versions(versions)
#   app.wormhole_got_message(data)
#   w.close()
#   app.wormhole_closed()
#
# * potential delegate options
#   wormhole(delegate=app, delegate_prefix="wormhole_",
#            delegate_args=(args, kwargs))


@attrs
@implementer(IWormhole)
class _DelegatedWormhole:
    _delegate = attrib()

    def __attrs_post_init__(self):
        self._key = None

    def _set_boss(self, boss):
        self._boss = boss

    # from above

    def allocate_code(self, code_length=2):
        self._boss.allocate_code(code_length)

    def input_code(self):
        return self._boss.input_code()

    def set_code(self, code):
        self._boss.set_code(code)

    # def serialize(self):
    #     s = {"serialized_wormhole_version": 1,
    #          "boss": self._boss.serialize(),
    #          }
    #     return s

    def send_message(self, plaintext):
        self._boss.send(plaintext)

    def derive_key(self, purpose, length):
        """Derive a new key from the established wormhole channel for some
        other purpose. This is a deterministic randomized function of the
        session key and the 'purpose' string (unicode/py3-string). This
        cannot be called until when_verifier() has fired, nor after close()
        was called.
        """
        if not isinstance(purpose, str):
            raise TypeError(type(purpose))
        if not self._key:
            raise NoKeyError()
        return derive_key(self._key, to_bytes(purpose), length)

    def close(self):
        self._boss.close()

    def debug_set_trace(self,
                        client_name,
                        which="B N M S O K SK R RC L C T",
                        file=sys.stderr):
        self._boss._set_trace(client_name, which, file)

    # from below
    def got_welcome(self, welcome):
        self._delegate.wormhole_got_welcome(welcome)

    def got_code(self, code):
        self._delegate.wormhole_got_code(code)

    def got_key(self, key):
        self._delegate.wormhole_got_unverified_key(key)
        self._key = key  # for derive_key()

    def got_verifier(self, verifier):
        self._delegate.wormhole_got_verifier(verifier)

    def got_versions(self, versions):
        self._delegate.wormhole_got_versions(versions)

    def received(self, plaintext):
        self._delegate.wormhole_got_message(plaintext)

    def closed(self, result):
        self._delegate.wormhole_closed(result)


@implementer(IWormhole, IDeferredWormhole)
class _DeferredWormhole:
    def __init__(self, reactor, eq, _enable_dilate=False):
        self._reactor = reactor
        self._welcome_observer = OneShotObserver(eq)
        self._code_observer = OneShotObserver(eq)
        self._key = None
        self._key_observer = OneShotObserver(eq)
        self._verifier_observer = OneShotObserver(eq)
        self._version_observer = OneShotObserver(eq)
        self._received_observer = SequenceObserver(eq)
        self._closed = False
        self._closed_observer = OneShotObserver(eq)

        self._enable_dilate = _enable_dilate

    def _set_boss(self, boss):
        self._boss = boss

    # from above
    def get_code(self):
        # TODO: consider throwing error unless one of allocate/set/input_code
        # was called first. It's legit to grab the Deferred before triggering
        # the process that will cause it to fire, but forbidding that
        # ordering would make it easier to cause programming errors that
        # forget to trigger it entirely.
        return self._code_observer.when_fired()

    def get_welcome(self):
        return self._welcome_observer.when_fired()

    def get_unverified_key(self):
        return self._key_observer.when_fired()

    def get_verifier(self):
        return self._verifier_observer.when_fired()

    def get_versions(self):
        return self._version_observer.when_fired()

    def get_message(self):
        return self._received_observer.when_next_event()

    def allocate_code(self, code_length=2):
        self._boss.allocate_code(code_length)

    def input_code(self):
        return self._boss.input_code()

    def set_code(self, code):
        self._boss.set_code(code)

    # no .serialize in Deferred-mode

    def send_message(self, plaintext):
        self._boss.send(plaintext)

    def derive_key(self, purpose, length):
        """Derive a new key from the established wormhole channel for some
        other purpose. This is a deterministic randomized function of the
        session key and the 'purpose' string (unicode/py3-string). This
        cannot be called until when_verified() has fired, nor after close()
        was called.
        """
        if not isinstance(purpose, str):
            raise TypeError(type(purpose))
        if not self._key:
            raise NoKeyError()
        return derive_key(self._key, to_bytes(purpose), length)

    # todo: note that only DeferredWormhole has this .. the
    # DelegatedWormhole thus has no way to dilate?

    # todo: transit_relay_locations (plural) probably, and ability to
    # pass a list? (there's a TODO about this is connector.py too)
    def dilate(self, transit_relay_location=None, no_listen=False, on_status_update=None, ping_interval=None, expected_subprotocols=None):
        """
        :returns DilatedWormhole: an instance for accessing dilation
            functionality. This includes creating endpoints that open
            new subchannels (i.e. the OPEN goes from us to the other
            peer).
        """
        if not self._enable_dilate:
            raise NotImplementedError
        return self._boss.dilate(transit_relay_location, no_listen, on_status_update, ping_interval, expected_subprotocols)

    def close(self):
        # fails with WormholeError unless we established a connection
        # (state=="happy"). Fails with WrongPasswordError (a subclass of
        # WormholeError) if state=="scary".
        d = self._closed_observer.when_fired()  # maybe Failure
        if not self._closed:
            self._boss.close()  # only need to close if it wasn't already
        return d

    def debug_set_trace(self,
                        client_name,
                        which="B N M S O K SK R RC L A I C T",
                        file=sys.stderr):
        self._boss._set_trace(client_name, which, file)

    # from below
    def got_welcome(self, welcome):
        self._welcome_observer.fire_if_not_fired(welcome)

    def got_code(self, code):
        self._code_observer.fire_if_not_fired(code)

    def got_key(self, key):
        self._key = key  # for derive_key()
        self._key_observer.fire_if_not_fired(key)

    def got_verifier(self, verifier):
        self._verifier_observer.fire_if_not_fired(verifier)

    def got_versions(self, versions):
        self._version_observer.fire_if_not_fired(versions)

    def received(self, plaintext):
        self._received_observer.fire(plaintext)

    def closed(self, result):
        self._closed = True
        # print("closed", result, type(result), file=sys.stderr)
        if isinstance(result, Exception):
            # everything pending gets an error, including close()
            f = failure.Failure(result)
            self._closed_observer.error(f)
        else:
            # everything pending except close() gets an error:
            # w.get_code()/welcome/unverified_key/verifier/versions/message
            f = failure.Failure(WormholeClosed(result))
            # but w.close() only gets error if we're unhappy
            self._closed_observer.fire_if_not_fired(result)
        self._welcome_observer.error(f)
        self._code_observer.error(f)
        self._key_observer.error(f)
        self._verifier_observer.error(f)
        self._version_observer.error(f)
        self._received_observer.fire(f)


def create(
        appid,
        relay_url,
        reactor,  # use keyword args for everything else
        versions={},
        delegate=None,
        journal=None,
        tor=None,
        timing=None,
        stderr=sys.stderr,
        dilation=None,
        _eventual_queue=None,
        on_status_update=None):
    timing = timing or DebugTiming()
    side = bytes_to_hexstr(os.urandom(5))
    journal = journal or ImmediateJournal()
    eq = _eventual_queue or EventualQueue(reactor)
    cooperator = Cooperator(scheduler=eq.eventually)

    if delegate:
        w = _DelegatedWormhole(delegate)
    else:
        w = _DeferredWormhole(reactor, eq, _enable_dilate=bool(dilation))
    # this indicates Wormhole capabilities
    wormhole_versions = {
        "can-dilate": DILATION_VERSIONS,
        "dilation-abilities": Connector.get_connection_abilities(),
    }

    if not dilation:
        wormhole_versions = {}
    wormhole_versions["app_versions"] = versions  # app-specific capabilities
    v = __version__
    if isinstance(v, bytes):
        v = v.decode("utf-8", errors="replace")
    client_version = ("python", v)
    b = Boss(w, side, relay_url, appid, wormhole_versions, client_version,
             reactor, eq, cooperator, journal, tor, timing, on_status_update)
    w._set_boss(b)
    b.start()
    return w


# def from_serialized(serialized, reactor, delegate,
#                     journal=None, tor=None,
#                     timing=None, stderr=sys.stderr):
#     assert serialized["serialized_wormhole_version"] == 1
#     timing = timing or DebugTiming()
#     w = _DelegatedWormhole(delegate)
#     # now unpack state machines, including the SPAKE2 in Key
#     b = Boss.from_serialized(w, serialized["boss"], reactor, journal, timing)
#     w._set_boss(b)
#     b.start() # ??
#     raise NotImplemented
#     # should the new Wormhole call got_code? only if it wasn't called before.



import contextlib

from zope.interface import implementer

from ._interfaces import IJournal


@implementer(IJournal)
class Journal:
    def __init__(self, save_checkpoint):
        self._save_checkpoint = save_checkpoint
        self._outbound_queue = []
        self._processing = False

    def queue_outbound(self, fn, *args, **kwargs):
        assert self._processing
        self._outbound_queue.append((fn, args, kwargs))

    @contextlib.contextmanager
    def process(self):
        assert not self._processing
        assert not self._outbound_queue
        self._processing = True
        yield  # process inbound messages, change state, queue outbound
        self._save_checkpoint()
        for (fn, args, kwargs) in self._outbound_queue:
            fn(*args, **kwargs)
        self._outbound_queue[:] = []
        self._processing = False


@implementer(IJournal)
class ImmediateJournal:
    def __init__(self):
        pass

    def queue_outbound(self, fn, *args, **kwargs):
        fn(*args, **kwargs)

    @contextlib.contextmanager
    def process(self):
        yield



from wormhole._dilation.manager import _find_shared_versions


def test_versions():
    assert _find_shared_versions([],[]) is None
    assert _find_shared_versions(["foo"],["foo"]) == "foo"
    assert _find_shared_versions(["foo","bar"],["foo"]) == "foo"
    assert _find_shared_versions(["bar","foo"],["foo","bar"]) == "bar"
    assert _find_shared_versions([],["foo","bar"]) is None
    assert _find_shared_versions(["foo","bar"],[]) is None



import gc
from unittest import mock
from zope.interface import alsoProvides
import twisted.logger
from twisted.internet.task import Clock
from twisted.internet.defer import Deferred
from twisted.internet.address import IPv4Address, IPv6Address, HostnameAddress
from twisted.internet.error import ConnectError
import pytest

from ...eventual import EventualQueue
from ..._interfaces import IDilationManager, IDilationConnector
from ..._hints import DirectTCPV1Hint, RelayV1Hint, TorTCPV1Hint
from ..._dilation import roles
from ..._dilation._noise import NoiseConnection
from ..._dilation.connection import KCM
from ..._dilation.connector import (Connector,
                                    build_sided_relay_handshake,
                                    build_noise,
                                    describe_inbound,
                                    OutboundConnectionFactory,
                                    InboundConnectionFactory,
                                    PROLOGUE_LEADER, PROLOGUE_FOLLOWER,
                                    )
from ..._status import DilationHint
from .common import clear_mock_calls


class Observer:
    """
    from: https://github.com/pytest-dev/pytest-twisted/issues/4
    with a few tweaks for Twisted changes over time
    """
    def __init__(self):
        self.failures = []

    def __call__(self, event_dict):
        is_error = event_dict.get('isError')
        s = 'Unhandled Error'.casefold()
        is_unhandled = s in event_dict.get('log_text', '').casefold()

        if is_error and is_unhandled:
            self.failures.append(event_dict)

    def assert_empty(self):
        assert [] == self.failures


@pytest.fixture
def assert_no_unhandled_errbacks():
    """
    A fixture which will fail the test if Observer believes that
    any unhandled errors were logged by Twisted.
    """
    observer = Observer()
    twisted.logger.globalLogPublisher.addObserver(observer)

    yield

    gc.collect()
    twisted.logger.globalLogPublisher.removeObserver(observer)

    observer.assert_empty()


def test_build():
    key = b"k"*32
    side = "12345678abcdabcd"
    assert build_sided_relay_handshake(key, side) == \
        b"please relay 3f4147851dbd2589d25b654ee9fb35ed0d3e5f19c5c5403e8e6a195c70f0577a" \
        b" for side 12345678abcdabcd\n"


def test_no_relay():
    c = mock.Mock()
    alsoProvides(c, IDilationConnector)
    p0 = mock.Mock()
    c.build_protocol = mock.Mock(return_value=p0)
    relay_handshake = None
    f = OutboundConnectionFactory(c, relay_handshake, "desc")
    addr = object()
    p = f.buildProtocol(addr)
    assert p is p0
    assert c.mock_calls == [mock.call.build_protocol(addr, "desc")]
    assert p.mock_calls == []
    assert p.factory is f


def test_with_relay():
    c = mock.Mock()
    alsoProvides(c, IDilationConnector)
    p0 = mock.Mock()
    c.build_protocol = mock.Mock(return_value=p0)
    relay_handshake = b"relay handshake"
    f = OutboundConnectionFactory(c, relay_handshake, "desc")
    addr = object()
    p = f.buildProtocol(addr)
    assert p is p0
    assert c.mock_calls == [mock.call.build_protocol(addr, "desc")]
    assert p.mock_calls == [mock.call.use_relay(relay_handshake)]
    assert p.factory is f


def test_build_inbound():
    c = mock.Mock()
    alsoProvides(c, IDilationConnector)
    p0 = mock.Mock()
    c.build_protocol = mock.Mock(return_value=p0)
    f = InboundConnectionFactory(c)
    addr = IPv4Address("TCP", "1.2.3.4", 55)
    p = f.buildProtocol(addr)
    assert p is p0
    assert c.mock_calls == [mock.call.build_protocol(addr, "<-tcp:1.2.3.4:55")]
    assert p.factory is f


def make_connector(listen=True, tor=False, relay=None, role=roles.LEADER):
    class Holder:
        pass
    h = Holder()
    h.dilation_key = b"key"
    h.relay = relay
    h.manager = mock.Mock()
    alsoProvides(h.manager, IDilationManager)
    h.clock = Clock()
    h.reactor = h.clock
    h.eq = EventualQueue(h.clock)
    h.tor = None
    if tor:
        h.tor = mock.Mock()
    timing = None
    h.side = "abcd1234abcd5678"
    h.role = role
    c = Connector(h.dilation_key, h.relay, h.manager, h.reactor, h.eq,
                  not listen, h.tor, timing, h.side, h.role)
    return c, h


def test_build_connector():
    c, h = make_connector()
    c, h = make_connector(relay="tcp:host:1234")


def test_connection_abilities():
    assert Connector.get_connection_abilities() == \
                     [{"type": "direct-tcp-v1"},
                      {"type": "relay-v1"},
                      ]


@pytest.mark.skipif(not NoiseConnection, reason="noiseprotocol required")
def test_build_noise():
    build_noise()


def test_build_protocol_leader():
    c, h = make_connector(role=roles.LEADER)
    n0 = mock.Mock()
    p0 = mock.Mock()
    addr = object()
    with mock.patch("wormhole._dilation.connector.build_noise",
                    return_value=n0) as bn:
        with mock.patch("wormhole._dilation.connector.DilatedConnectionProtocol",
                        return_value=p0) as dcp:
            p = c.build_protocol(addr, "desc")
    assert bn.mock_calls == [mock.call()]
    assert n0.mock_calls == [mock.call.set_psks(h.dilation_key),
                                     mock.call.set_as_initiator()]
    assert p is p0
    assert dcp.mock_calls == \
                     [mock.call(h.eq, h.role, "desc", c, n0,
                                PROLOGUE_LEADER, PROLOGUE_FOLLOWER)]

def test_build_protocol_follower():
    c, h = make_connector(role=roles.FOLLOWER)
    n0 = mock.Mock()
    p0 = mock.Mock()
    addr = object()
    with mock.patch("wormhole._dilation.connector.build_noise",
                    return_value=n0) as bn:
        with mock.patch("wormhole._dilation.connector.DilatedConnectionProtocol",
                        return_value=p0) as dcp:
            p = c.build_protocol(addr, "desc")
    assert bn.mock_calls == [mock.call()]
    assert n0.mock_calls == [mock.call.set_psks(h.dilation_key),
                                     mock.call.set_as_responder()]
    assert p is p0
    assert dcp.mock_calls == \
                     [mock.call(h.eq, h.role, "desc", c, n0,
                                PROLOGUE_FOLLOWER, PROLOGUE_LEADER)]

def test_start_stop():
    c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
    c.start()
    # no relays, so it publishes no hints
    assert h.manager.mock_calls == []
    # and no listener, so nothing happens until we provide a hint
    c.stop()
    # we stop while we're connecting, so no connections must be stopped

def test_empty():
    c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()
    # no relays, so it publishes no hints
    assert h.manager.mock_calls == []
    # and no listener, so nothing happens until we provide a hint
    assert c._schedule_connection.mock_calls == []
    c.stop()

def test_basic():
    c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()
    # no relays, so it publishes no hints
    assert h.manager.mock_calls == []
    # and no listener, so nothing happens until we provide a hint
    assert c._schedule_connection.mock_calls == []

    hint = DirectTCPV1Hint("foo", 55, 0.0)
    c.got_hints([hint])

    # received hints don't get published
    assert h.manager.mock_calls == [mock.call._hint_status([DilationHint(url='foo:55', is_direct=True)])]
    # they just schedule a connection
    assert c._schedule_connection.mock_calls == \
                     [mock.call(0.0, DirectTCPV1Hint("foo", 55, 0.0),
                                is_relay=False)]


def test_connection_error(assert_no_unhandled_errbacks):
    """
    An error is triggered on our connection attempt
    """
    c, h = make_connector(listen=False, relay=None, role=roles.LEADER)
    # _schedule_connection does a delayed call on _connect()
    d_error = Deferred()
    d_error.errback(ConnectError("no good reason"))
    c._connect = mock.Mock(return_value=d_error)
    c.start()
    # no relays, so it publishes no hints
    assert h.manager.mock_calls == []

    hint = DirectTCPV1Hint("foo", 55, 0.0)
    c.got_hints([hint])

    # received hints don't get published
    assert h.manager.mock_calls == [mock.call._hint_status([DilationHint(url='foo:55', is_direct=True)])]
    # the "_connect()" call is delayed, almost certainly by "0.0"
    # seconds in this case, but we need to trigger a reactor turn
    h.clock.advance(1)
    assert len(c._connect.mock_calls) == 1
    # without the fix in
    # https://github.com/magic-wormhole/magic-wormhole/pull/670 the
    # "no_unhandled_errors" fixture will assert


def test_listen_addresses():
    c, h = make_connector(listen=True, role=roles.LEADER)
    with mock.patch("wormhole.ipaddrs.find_addresses",
                    return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]):
        assert c._get_listener_addresses() == \
                         ["1.2.3.4", "5.6.7.8"]
    with mock.patch("wormhole.ipaddrs.find_addresses",
                    return_value=["127.0.0.1"]):
        # some test hosts, including the appveyor VMs, *only* have
        # 127.0.0.1, and the tests will hang badly if we remove it.
        assert c._get_listener_addresses() == ["127.0.0.1"]

def test_listen():
    c, h = make_connector(listen=True, role=roles.LEADER)
    c._start_listener = mock.Mock()
    with mock.patch("wormhole.ipaddrs.find_addresses",
                    return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]):
        c.start()
    assert c._start_listener.mock_calls == \
                     [mock.call(["1.2.3.4", "5.6.7.8"])]

def test_start_listen():
    c, h = make_connector(listen=True, role=roles.LEADER)
    ep = mock.Mock()
    d = Deferred()
    ep.listen = mock.Mock(return_value=d)
    with mock.patch("wormhole._dilation.connector.serverFromString",
                    return_value=ep) as sfs:
        c._start_listener(["1.2.3.4", "5.6.7.8"])
    assert sfs.mock_calls == [mock.call(h.reactor, "tcp:0")]
    lp = mock.Mock()
    host = mock.Mock()
    host.port = 66
    lp.getHost = mock.Mock(return_value=host)
    d.callback(lp)
    assert h.manager.mock_calls == \
                     [mock.call.send_hints([{"type": "direct-tcp-v1",
                                             "hostname": "1.2.3.4",
                                             "port": 66,
                                             "priority": 0.0
                                             },
                                            {"type": "direct-tcp-v1",
                                             "hostname": "5.6.7.8",
                                             "port": 66,
                                             "priority": 0.0
                                             },
                                            ])]

def test_schedule_connection_no_relay():
    c, h = make_connector(listen=True, role=roles.LEADER)
    hint = DirectTCPV1Hint("foo", 55, 0.0)
    ep = mock.Mock()
    with mock.patch("wormhole._dilation.connector.endpoint_from_hint_obj",
                    side_effect=[ep]) as efho:
        c._schedule_connection(0.0, hint, False)
    assert efho.mock_calls == [mock.call(hint, h.tor, h.reactor)]
    assert ep.mock_calls == []
    d = Deferred()
    ep.connect = mock.Mock(side_effect=[d])
    # direct hints are scheduled for T+0.0
    f = mock.Mock()
    with mock.patch("wormhole._dilation.connector.OutboundConnectionFactory",
                    return_value=f) as ocf:
        h.clock.advance(1.0)
    assert ocf.mock_calls == [mock.call(c, None, "->tcp:foo:55")]
    assert ep.connect.mock_calls == [mock.call(f)]
    p = mock.Mock()
    d.callback(p)
    assert p.mock_calls == \
                     [mock.call.when_disconnected(),
                      mock.call.when_disconnected().addCallback(c._pending_connections.discard)]

def test_schedule_connection_relay():
    c, h = make_connector(listen=True, role=roles.LEADER)
    hint = DirectTCPV1Hint("foo", 55, 0.0)
    ep = mock.Mock()
    with mock.patch("wormhole._dilation.connector.endpoint_from_hint_obj",
                    side_effect=[ep]) as efho:
        c._schedule_connection(0.0, hint, True)
    assert efho.mock_calls == [mock.call(hint, h.tor, h.reactor)]
    assert ep.mock_calls == []
    d = Deferred()
    ep.connect = mock.Mock(side_effect=[d])
    # direct hints are scheduled for T+0.0
    f = mock.Mock()
    with mock.patch("wormhole._dilation.connector.OutboundConnectionFactory",
                    return_value=f) as ocf:
        h.clock.advance(1.0)
    handshake = build_sided_relay_handshake(h.dilation_key, h.side)
    assert ocf.mock_calls == [mock.call(c, handshake, "->relay:tcp:foo:55")]

def test_listen_but_tor():
    c, h = make_connector(listen=True, tor=True, role=roles.LEADER)
    with mock.patch("wormhole.ipaddrs.find_addresses",
                    return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]) as fa:
        c.start()
    # don't even look up addresses
    assert fa.mock_calls == []
    # no relays and the listener isn't ready yet, so no hints yet
    assert h.manager.mock_calls == []

def test_no_listen():
    c, h = make_connector(listen=False, tor=False, role=roles.LEADER)
    with mock.patch("wormhole.ipaddrs.find_addresses",
                    return_value=["127.0.0.1", "1.2.3.4", "5.6.7.8"]) as fa:
        c.start()
    # don't even look up addresses
    assert fa.mock_calls == []
    assert h.manager.mock_calls == []

def test_relay_delay():
    # given a direct connection and a relay, we should see the direct
    # connection initiated at T+0 seconds, and the relay at T+RELAY_DELAY
    c, h = make_connector(listen=True, relay=None, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c._start_listener = mock.Mock()
    c.start()
    hint1 = DirectTCPV1Hint("foo", 55, 0.0)
    hint2 = DirectTCPV1Hint("bar", 55, 0.0)
    hint3 = RelayV1Hint([DirectTCPV1Hint("relay", 55, 0.0)])
    c.got_hints([hint1, hint2, hint3])
    assert c._schedule_connection.mock_calls == \
                     [mock.call(0.0, hint1, is_relay=False),
                      mock.call(0.0, hint2, is_relay=False),
                      mock.call(c.RELAY_DELAY, hint3.hints[0], is_relay=True),
                      ]

def test_initial_relay():
    c, h = make_connector(listen=False, relay="tcp:foo:55", role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()
    assert h.manager.mock_calls == [
        mock.call.send_hints([
            {
                "type": "relay-v1",
                "hints": [
                    {
                        "type": "direct-tcp-v1",
                        "hostname": "foo",
                        "port": 55,
                        "priority": 0.0
                    },
                ],
            },
        ]),
        mock.call._hint_status([DilationHint(url='foo:55', is_direct=False)]),
    ]
    assert c._schedule_connection.mock_calls == \
                     [mock.call(0.0, DirectTCPV1Hint("foo", 55, 0.0),
                                is_relay=True)]

def test_tor_no_manager():
    # tor hints should be ignored if we don't have a Tor manager to use them
    c, h = make_connector(listen=False, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()
    hint = TorTCPV1Hint("foo", 55, 0.0)
    c.got_hints([hint])
    assert h.manager.mock_calls == [
        mock.call._hint_status([]),
    ]
    assert c._schedule_connection.mock_calls == []

def test_tor_with_manager():
    # tor hints should be processed if we do have a Tor manager
    c, h = make_connector(listen=False, tor=True, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()
    hint = TorTCPV1Hint("foo", 55, 0.0)
    c.got_hints([hint])
    assert c._schedule_connection.mock_calls == \
                     [mock.call(0.0, hint, is_relay=False)]

def test_priorities():
    # given two hints with different priorities, we should somehow prefer
    # one. This is a placeholder to fill in once we implement priorities.
    pass


def test_one_leader():
    c, h = make_connector(listen=True, role=roles.LEADER)
    lp = mock.Mock()

    def start_listener(addresses):
        c._listeners.add(lp)
    c._start_listener = start_listener
    c._schedule_connection = mock.Mock()
    c.start()
    assert c._listeners == {lp}

    p1 = mock.Mock()  # DilatedConnectionProtocol instance
    c.add_candidate(p1)
    assert h.manager.mock_calls == []
    h.eq.flush_sync()
    assert h.manager.mock_calls == [mock.call.connector_connection_made(p1)]
    assert p1.mock_calls == \
                     [mock.call.select(h.manager),
                      mock.call.send_record(KCM())]
    assert lp.mock_calls[0] == mock.call.stopListening()
    # stop_listeners() uses a DeferredList, so we ignore the second call

def test_one_follower():
    c, h = make_connector(listen=True, role=roles.FOLLOWER)
    lp = mock.Mock()

    def start_listener(addresses):
        c._listeners.add(lp)
    c._start_listener = start_listener
    c._schedule_connection = mock.Mock()
    c.start()
    assert c._listeners == {lp}

    p1 = mock.Mock()  # DilatedConnectionProtocol instance
    c.add_candidate(p1)
    assert h.manager.mock_calls == []
    h.eq.flush_sync()
    assert h.manager.mock_calls == [mock.call.connector_connection_made(p1)]
    # just like LEADER, but follower doesn't send KCM now (it sent one
    # earlier, to tell the leader that this connection looks viable)
    assert p1.mock_calls == \
                     [mock.call.select(h.manager)]
    assert lp.mock_calls[0] == mock.call.stopListening()
    # stop_listeners() uses a DeferredList, so we ignore the second call

# TODO: make sure a pending connection is abandoned when the listener
# answers successfully

# TODO: make sure a second pending connection is abandoned when the first
# connection succeeds

def test_late():
    c, h = make_connector(listen=False, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()

    p1 = mock.Mock()  # DilatedConnectionProtocol instance
    c.add_candidate(p1)
    assert h.manager.mock_calls == []
    h.eq.flush_sync()
    assert h.manager.mock_calls == [mock.call.connector_connection_made(p1)]
    clear_mock_calls(h.manager)
    assert p1.mock_calls == \
                     [mock.call.select(h.manager),
                      mock.call.send_record(KCM())]

    # late connection is ignored
    p2 = mock.Mock()
    c.add_candidate(p2)
    assert h.manager.mock_calls == []

# make sure an established connection is dropped when stop() is called
def test_stop():
    c, h = make_connector(listen=False, role=roles.LEADER)
    c._schedule_connection = mock.Mock()
    c.start()

    p1 = mock.Mock()  # DilatedConnectionProtocol instance
    c.add_candidate(p1)
    assert h.manager.mock_calls == []
    h.eq.flush_sync()
    assert p1.mock_calls == \
                     [mock.call.select(h.manager),
                      mock.call.send_record(KCM())]
    assert h.manager.mock_calls == [mock.call.connector_connection_made(p1)]

    c.stop()


def test_describe_inbound():
    assert describe_inbound(HostnameAddress("example.com", 1234)) == \
                     "<-tcp:example.com:1234"
    assert describe_inbound(IPv4Address("TCP", "1.2.3.4", 1234)) == \
                     "<-tcp:1.2.3.4:1234"
    assert describe_inbound(IPv6Address("TCP", "::1", 1234)) == \
                     "<-tcp:[::1]:1234"
    other = "none-of-the-above"
    assert describe_inbound(other) == f"<-{other!r}"



from collections import namedtuple
from attr import attrs, attrib
from attr.validators import instance_of
from automat import MethodicalMachine
from zope.interface import Interface, implementer
from twisted.python import log
from twisted.internet.protocol import Protocol
from twisted.internet.interfaces import ITransport
from .._interfaces import IDilationConnector
from ..observer import OneShotObserver
from ..util import provides
from .encode import to_be4, from_be4
from .roles import LEADER, FOLLOWER
from ._noise import NoiseInvalidMessage, NoiseHandshakeError, NOISE_MAX_PAYLOAD, NOISE_MAX_CIPHERTEXT

# InboundFraming is given data and returns Frames (Noise wire-side
# bytestrings). It handles the relay handshake and the prologue. The Frames it
# returns are either the ephemeral key (the Noise "handshake") or ciphertext
# messages.

# The next object up knows whether it's expecting a Handshake or a message. It
# feeds the first into Noise as a handshake, it feeds the rest into Noise as a
# message (which produces a plaintext stream). It emits tokens that are either
# "i've finished with the handshake (so you can send the KCM if you want)", or
# "here is a decrypted message (which might be the KCM)".

# the transmit direction goes directly to transport.write, and doesn't touch
# the state machine. we can do this because the way we encode/encrypt/frame
# things doesn't depend upon the receiver state. It would be more safe to e.g.
# prohibit sending ciphertext frames unless we're in the received-handshake
# state, but then we'll be in the middle of an inbound state transition ("we
# just received the handshake, so you can send a KCM now") when we perform an
# operation that depends upon the state (send_plaintext(kcm)), which is not a
# coherent/safe place to touch the state machine.

# we could set a flag and test it from inside send_plaintext, which kind of
# violates the state machine owning the state (ideally all "if" statements
# would be translated into same-input transitions from different starting
# states). For the specific question of sending plaintext frames, Noise will
# refuse us unless it's ready anyways, so the question is probably moot.


class IFramer(Interface):
    pass


class IRecord(Interface):
    pass


def first(seq):
    return seq[0]


class Disconnect(Exception):
    pass

# all connections look like:
# (step 1: only for outbound connections)
# 1: if we're connecting to a transit relay:
#    * send "sided relay handshake": "please relay TOKEN for side SIDE\n"
#    * the relay will send "ok\n" if/when our peer connects
#    * a non-relay will probably send junk
#    * wait for "ok\n", hang up if we get anything different
# (all subsequent steps are for both inbound and outbound connections)
# 2: send PROLOGUE_LEADER/FOLLOWER: "Magic-Wormhole Dilation Handshake v1 (l/f)\n\n"
# 3: wait for the opposite PROLOGUE string, else hang up
# (everything past this point is a Frame, with be4 length prefix. Frames are
#  either noise handshake or an encrypted message)
# 4: if LEADER, send noise handshake string. if FOLLOWER, wait for it
#    LEADER: m=n.write_message(), FOLLOWER: n.read_message(m)
# 5: if FOLLOWER, send noise response string. if LEADER, wait for it
#    FOLLOWER: m=n.write_message(), LEADER: n.read_message(m)
# 6: if FOLLOWER: send KCM (m=n.encrypt('')), wait for KCM (n.decrypt(m))
#    if LEADER: wait for KCM, gather viable connections, select
#               send KCM over selected connection, drop the rest
# 7: both: send Ping/Pong/Open/Data/Close/Ack records (n.encrypt(rec))


RelayOK = namedtuple("RelayOk", [])
Prologue = namedtuple("Prologue", [])
Frame = namedtuple("Frame", ["frame"])


@attrs
@implementer(IFramer)
class _Framer:
    _transport = attrib(validator=provides(ITransport))
    _outbound_prologue = attrib(validator=instance_of(bytes))
    _inbound_prologue = attrib(validator=instance_of(bytes))
    _buffer = b""
    _can_send_frames = False

    # in: use_relay
    # in: connectionMade, dataReceived
    # out: prologue_received, frame_received
    # out (shared): transport.loseConnection
    # out (shared): transport.write (relay handshake, prologue)
    # states: want_relay, want_prologue, want_frame
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)  # pragma: no cover

    @m.state()
    def want_relay(self):
        pass  # pragma: no cover

    @m.state(initial=True)
    def want_prologue(self):
        pass  # pragma: no cover

    @m.state()
    def want_frame(self):
        pass  # pragma: no cover

    @m.input()
    def use_relay(self, relay_handshake):
        pass

    @m.input()
    def connectionMade(self):
        pass

    @m.input()
    def parse(self):
        pass

    @m.input()
    def got_relay_ok(self):
        pass

    @m.input()
    def got_prologue(self):
        pass

    @m.output()
    def store_relay_handshake(self, relay_handshake):
        self._outbound_relay_handshake = relay_handshake
        self._expected_relay_handshake = b"ok\n"  # TODO: make this configurable

    @m.output()
    def send_relay_handshake(self):
        self._transport.write(self._outbound_relay_handshake)

    @m.output()
    def send_prologue(self):
        self._transport.write(self._outbound_prologue)

    @m.output()
    def parse_relay_ok(self):
        if self._get_expected("relay_ok", self._expected_relay_handshake):
            return RelayOK()

    @m.output()
    def parse_prologue(self):
        if self._get_expected("prologue", self._inbound_prologue):
            return Prologue()

    @m.output()
    def can_send_frames(self):
        self._can_send_frames = True  # for assertion in send_frame()

    @m.output()
    def parse_frame(self):
        if len(self._buffer) < 4:
            return None
        frame_length = from_be4(self._buffer[0:4])
        if len(self._buffer) < 4 + frame_length:
            return None
        frame = self._buffer[4:4 + frame_length]
        self._buffer = self._buffer[4 + frame_length:]  # TODO: avoid copy
        return Frame(frame=frame)

    want_prologue.upon(use_relay, outputs=[store_relay_handshake],
                       enter=want_relay)

    want_relay.upon(connectionMade, outputs=[send_relay_handshake],
                    enter=want_relay)
    want_relay.upon(parse, outputs=[parse_relay_ok], enter=want_relay,
                    collector=first)
    want_relay.upon(got_relay_ok, outputs=[send_prologue], enter=want_prologue)

    want_prologue.upon(connectionMade, outputs=[send_prologue],
                       enter=want_prologue)
    want_prologue.upon(parse, outputs=[parse_prologue], enter=want_prologue,
                       collector=first)
    want_prologue.upon(got_prologue, outputs=[can_send_frames], enter=want_frame)

    want_frame.upon(parse, outputs=[parse_frame], enter=want_frame,
                    collector=first)

    def _get_expected(self, name, expected):
        lb = len(self._buffer)
        le = len(expected)
        if self._buffer.startswith(expected):
            # if the buffer starts with the expected string, consume it and
            # return True
            self._buffer = self._buffer[le:]
            return True
        if not expected.startswith(self._buffer):
            # we're not on track: the data we've received so far does not
            # match the expected value, so this can't possibly be right.
            # Don't complain until we see the expected length, or a newline,
            # so we can capture the weird input in the log for debugging.
            if (b"\n" in self._buffer or lb >= le):
                log.msg(f"bad {name}: {self._buffer[:le]}")
                raise Disconnect()
            return False  # wait a bit longer
        # good so far, just waiting for the rest
        return False

    # external API is: connectionMade, add_and_parse, and send_frame

    def add_and_parse(self, data):
        # we can't make this an @m.input because we can't change the state
        # from within an input. Instead, let the state choose the parser to
        # use, then use the parsed token to drive a state transition.
        self._buffer += data
        while True:
            # it'd be nice to use an iterator here, but since self.parse()
            # dispatches to a different parser (depending upon the current
            # state), we'd be using multiple iterators
            token = self.parse()
            if isinstance(token, RelayOK):
                self.got_relay_ok()
            elif isinstance(token, Prologue):
                self.got_prologue()
                yield token  # triggers send_handshake
            elif isinstance(token, Frame):
                yield token
            else:
                break

    def send_frame(self, frame):
        assert self._can_send_frames
        self._transport.write(to_be4(len(frame)) + frame)

# RelayOK: Newline-terminated buddy-is-connected response from Relay.
#          First data received from relay.
# Prologue: double-newline-terminated this-is-really-wormhole response
#           from peer. First data received from peer.
# Frame: Either handshake or encrypted message. Length-prefixed on wire.
# Handshake: the Noise ephemeral key, first framed message
# Message: plaintext: encoded KCM/PING/PONG/OPEN/DATA/CLOSE/ACK
# KCM: Key Confirmation Message (encrypted b"\x00"). First frame
#      from peer. Sent immediately by Follower, after Selection by Leader.
# Record: namedtuple of KCM/Open/Data/Close/Ack/Ping/Pong


Handshake = namedtuple("Handshake", [])
# decrypted frames: produces KCM, Ping, Pong, Open, Data, Close, Ack
KCM = namedtuple("KCM", [])
Ping = namedtuple("Ping", ["ping_id"])  # ping_id is arbitrary 4-byte value
Pong = namedtuple("Pong", ["ping_id"])
Open = namedtuple("Open", ["seqnum", "scid", "subprotocol"])  # seqnum is integer, subprotocol is str
Data = namedtuple("Data", ["seqnum", "scid", "data"])
Close = namedtuple("Close", ["seqnum", "scid"])  # scid is integer
Ack = namedtuple("Ack", ["resp_seqnum"])  # resp_seqnum is integer
Records = (KCM, Ping, Pong, Open, Data, Close, Ack)
Handshake_or_Records = (Handshake,) + Records

T_KCM = b"\x00"
T_PING = b"\x01"
T_PONG = b"\x02"
T_OPEN = b"\x03"
T_DATA = b"\x04"
T_CLOSE = b"\x05"
T_ACK = b"\x06"


def parse_record(plaintext):
    msgtype = plaintext[0:1]
    if msgtype == T_KCM:
        return KCM()
    if msgtype == T_PING:
        ping_id = plaintext[1:5]
        return Ping(ping_id)
    if msgtype == T_PONG:
        ping_id = plaintext[1:5]
        return Pong(ping_id)
    if msgtype == T_OPEN:
        scid = from_be4(plaintext[1:5])
        seqnum = from_be4(plaintext[5:9])
        subprotocol = str(plaintext[9:], "utf8")
        return Open(seqnum, scid, subprotocol)
    if msgtype == T_DATA:
        scid = from_be4(plaintext[1:5])
        seqnum = from_be4(plaintext[5:9])
        data = plaintext[9:]
        return Data(seqnum, scid, data)
    if msgtype == T_CLOSE:
        scid = from_be4(plaintext[1:5])
        seqnum = from_be4(plaintext[5:9])
        return Close(seqnum, scid)
    if msgtype == T_ACK:
        resp_seqnum = from_be4(plaintext[1:5])
        return Ack(resp_seqnum)
    log.err(f"received unknown message type: {plaintext}")
    raise ValueError()


def encode_record(r):
    if isinstance(r, KCM):
        return T_KCM
    if isinstance(r, Ping):
        return T_PING + r.ping_id
    if isinstance(r, Pong):
        return T_PONG + r.ping_id
    if isinstance(r, Open):
        assert isinstance(r.scid, int)
        assert isinstance(r.seqnum, int)
        assert isinstance(r.subprotocol, str)
        return T_OPEN + to_be4(r.scid) + to_be4(r.seqnum) + r.subprotocol.encode("utf8")
    if isinstance(r, Data):
        assert isinstance(r.scid, int)
        assert isinstance(r.seqnum, int)
        return T_DATA + to_be4(r.scid) + to_be4(r.seqnum) + r.data
    if isinstance(r, Close):
        assert isinstance(r.scid, int)
        assert isinstance(r.seqnum, int)
        return T_CLOSE + to_be4(r.scid) + to_be4(r.seqnum)
    if isinstance(r, Ack):
        assert isinstance(r.resp_seqnum, int)
        return T_ACK + to_be4(r.resp_seqnum)
    raise TypeError(r)


def _is_role(_record, _attr, value):
    if value not in [LEADER, FOLLOWER]:
        raise ValueError("role must be LEADER or FOLLOWER")


@attrs
@implementer(IRecord)
class _Record:
    _framer = attrib(validator=provides(IFramer))
    _noise = attrib()

    _role = attrib(default="unspecified", validator=_is_role)  # for debugging

    n = MethodicalMachine()
    # TODO: set_trace

    def __attrs_post_init__(self):
        self._noise.start_handshake()

    # in: role=
    # in: prologue_received, frame_received
    # out: handshake_received, record_received
    # out: transport.write (noise handshake, encrypted records)
    # states: want_prologue, want_handshake, want_record

    @n.state(initial=True)
    def no_role_set(self):
        pass  # pragma: no cover

    @n.state()
    def want_prologue_leader(self):
        pass  # pragma: no cover

    @n.state()
    def want_prologue_follower(self):
        pass  # pragma: no cover

    @n.state()
    def want_handshake_leader(self):
        pass  # pragma: no cover

    @n.state()
    def want_handshake_follower(self):
        pass  # pragma: no cover

    @n.state()
    def want_message(self):
        pass  # pragma: no cover

    @n.input()
    def set_role_leader(self):
        pass

    @n.input()
    def set_role_follower(self):
        pass

    @n.input()
    def got_prologue(self):
        pass

    @n.input()
    def got_frame(self, frame):
        pass

    @n.output()
    def ignore_and_send_handshake(self, frame):
        self._send_handshake()

    @n.output()
    def send_handshake(self):
        self._send_handshake()

    def _send_handshake(self):
        try:
            handshake = self._noise.write_message()  # generate the ephemeral key
        except NoiseHandshakeError as e:
            log.err(e, "noise error during handshake")
            raise
        self._framer.send_frame(handshake)

    @n.output()
    def process_handshake(self, frame):
        try:
            payload = self._noise.read_message(frame)
            # Noise can include unencrypted data in the handshake, but we don't
            # use it
            del payload
        except NoiseInvalidMessage as e:
            log.err(e, "bad inbound noise handshake")
            raise Disconnect()
        return Handshake()

    @n.output()
    def decrypt_message(self, frame):
        # opposite of the encoding: if we have _more_ than what a
        # single Noise packet can hold, we have to build up the real
        # plaintext incrementally.
        size = len(frame)
        try:
            if size <= NOISE_MAX_CIPHERTEXT:
                message = self._noise.decrypt(frame)
            else:
                start = 0
                message = b""
                while start < size:
                    ciphertext = frame[start:start + NOISE_MAX_CIPHERTEXT]
                    message += self._noise.decrypt(ciphertext)
                    start += NOISE_MAX_CIPHERTEXT
        except NoiseInvalidMessage as e:
            # if this happens during tests, flunk the test
            log.err(e, "bad inbound noise frame")
            raise Disconnect()
        return parse_record(message)

    no_role_set.upon(set_role_leader, outputs=[], enter=want_prologue_leader)
    want_prologue_leader.upon(got_prologue, outputs=[send_handshake],
                              enter=want_handshake_leader)
    want_handshake_leader.upon(got_frame, outputs=[process_handshake],
                               collector=first, enter=want_message)

    no_role_set.upon(set_role_follower, outputs=[], enter=want_prologue_follower)
    want_prologue_follower.upon(got_prologue, outputs=[],
                                enter=want_handshake_follower)
    want_handshake_follower.upon(got_frame, outputs=[process_handshake,
                                                     ignore_and_send_handshake],
                                 collector=first, enter=want_message)

    want_message.upon(got_frame, outputs=[decrypt_message],
                      collector=first, enter=want_message)

    # external API is: connectionMade, dataReceived, send_record

    def connectionMade(self):
        self._framer.connectionMade()

    def add_and_unframe(self, data):
        for token in self._framer.add_and_parse(data):
            if isinstance(token, Prologue):
                self.got_prologue()  # triggers send_handshake
            else:
                assert isinstance(token, Frame)
                yield self.got_frame(token.frame)  # Handshake or a Record type

    def send_record(self, r):
        message = encode_record(r)
        if len(message) <= NOISE_MAX_PAYLOAD:
            frame = self._noise.encrypt(message)
        else:
            # we want to put all the encrypted bytes into one "frame",
            # but there are more bytes than we can fit in a Noise
            # message .. so we chop them up
            start = 0
            frame = b""
            while start < len(message):
                this_msg = message[start:start + NOISE_MAX_PAYLOAD]
                cip = self._noise.encrypt(this_msg)
                frame += cip
                start += NOISE_MAX_PAYLOAD
        self._framer.send_frame(frame)


@attrs(eq=False)
class DilatedConnectionProtocol(Protocol):
    """I manage an L2 connection.

    When a new L2 connection is needed (as determined by the Leader),
    both Leader and Follower will initiate many simultaneous connections
    (probably TCP, but conceivably others). A subset will actually
    connect. A subset of those will successfully pass negotiation by
    exchanging handshakes to demonstrate knowledge of the session key.
    One of the negotiated connections will be selected by the Leader for
    active use, and the others will be dropped.

    At any given time, there is at most one active L2 connection.
    """

    _eventual_queue = attrib(repr=False)
    _role = attrib()
    _description = attrib()
    _connector = attrib(validator=provides(IDilationConnector), repr=False)
    _noise = attrib(repr=False)
    _outbound_prologue = attrib(validator=instance_of(bytes), repr=False)
    _inbound_prologue = attrib(validator=instance_of(bytes), repr=False)

    _use_relay = False
    _relay_handshake = None

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._manager = None  # set if/when we are selected
        self._disconnected = OneShotObserver(self._eventual_queue)
        self._can_send_records = False
        self._inbound_record_queue = []

    @m.state(initial=True)
    def unselected(self):
        pass  # pragma: no cover

    @m.state()
    def selecting(self):
        pass  # pragma: no cover

    @m.state()
    def selected(self):
        pass  # pragma: no cover

    @m.input()
    def got_kcm(self):
        pass

    @m.input()
    def select(self, manager):
        pass  # fires set_manager()

    @m.input()
    def got_record(self, record):
        pass

    @m.output()
    def add_candidate(self):
        self._connector.add_candidate(self)

    @m.output()
    def queue_inbound_record(self, record):
        # the Follower will see a dataReceived chunk containing both the KCM
        # (leader says we've been picked) and the first record.
        # Connector.consider takes an eventual-turn to decide to accept this
        # connection, which means the record will arrive before we get
        # .select() and move to the 'selected' state where we can
        # deliver_record. So we need to queue the record for a turn. TODO:
        # when we move to the sans-io event-driven scheme, this queue
        # shouldn't be necessary
        self._inbound_record_queue.append(record)

    @m.output()
    def set_manager(self, manager):
        self._manager = manager
        self.when_disconnected().addCallback(lambda c:
                                             manager.connector_connection_lost())

    @m.output()
    def send_status_have_peer(self, manager):
        assert self._manager is not None, "_manager must be set by now"
        self._manager.have_peer(self)

    @m.output()
    def can_send_records(self, manager):
        self._can_send_records = True

    @m.output()
    def process_inbound_queue(self, manager):
        while self._inbound_record_queue:
            r = self._inbound_record_queue.pop(0)
            self._manager.got_record(r)

    @m.output()
    def deliver_record(self, record):
        self._manager.got_record(record)

    unselected.upon(got_kcm, outputs=[add_candidate], enter=selecting)
    selecting.upon(got_record, outputs=[queue_inbound_record], enter=selecting)
    selecting.upon(select,
                   outputs=[set_manager, send_status_have_peer, can_send_records, process_inbound_queue],
                   enter=selected)
    selected.upon(got_record, outputs=[deliver_record], enter=selected)

    # called by Connector

    def use_relay(self, relay_handshake):
        assert isinstance(relay_handshake, bytes)
        self._use_relay = True
        self._relay_handshake = relay_handshake

    def when_disconnected(self):
        return self._disconnected.when_fired()

    def disconnect(self):
        self.transport.loseConnection()

    # select() called by Connector

    # called by Manager
    def send_record(self, record):
        assert self._can_send_records
        self._record.send_record(record)

    # IProtocol methods

    def connectionMade(self):
        try:
            framer = _Framer(self.transport,
                             self._outbound_prologue, self._inbound_prologue)
            if self._use_relay:
                framer.use_relay(self._relay_handshake)
            self._record = _Record(framer, self._noise, self._role)
            if self._role is LEADER:
                self._record.set_role_leader()
            else:
                self._record.set_role_follower()
            self._record.connectionMade()

        except:  # noqa
            log.err()
            raise

    def dataReceived(self, data):
        try:
            for token in self._record.add_and_unframe(data):
                assert isinstance(token, Handshake_or_Records)
                if isinstance(token, Handshake):
                    if self._role is FOLLOWER:
                        self._record.send_record(KCM())
                elif isinstance(token, KCM):
                    # if we're the leader, add this connection as a candidate.
                    # if we're the follower, accept this connection.
                    self.got_kcm()  # connector.add_candidate()
                else:
                    self.got_record(token)  # manager.got_record()
        except Disconnect:
            self.transport.loseConnection()

    def connectionLost(self, why=None):
        self._disconnected.fire(self)



import io
from binascii import hexlify, unhexlify

from nacl.exceptions import CryptoError
from nacl.secret import SecretBox
from twisted.internet import address, defer, endpoints, error, protocol, task
from twisted.internet.defer import gatherResults
from twisted.test import proto_helpers

from unittest import mock
from wormhole_transit_relay import transit_server
from pytest_twisted import ensureDeferred

from .. import transit
from .._hints import DirectTCPV1Hint
from ..errors import InternalError
from ..util import HKDF
import pytest


@ensureDeferred
async def test_one_winner():
    cancelled = set()
    contenders = [
        defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(5)
    ]
    d = transit.there_can_be_only_one(contenders)
    assert not d.called
    contenders[0].errback(ValueError())
    assert not d.called
    contenders[1].errback(TypeError())
    assert not d.called
    contenders[2].callback("yay")
    assert await d == "yay"
    assert cancelled == {3, 4}


@ensureDeferred
async def test_there_might_also_be_none():
    cancelled = set()
    contenders = [
        defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(4)
    ]
    d = transit.there_can_be_only_one(contenders)
    assert not d.called
    contenders[0].errback(ValueError())
    assert not d.called
    contenders[1].errback(TypeError())
    assert not d.called
    contenders[2].errback(TypeError())
    assert not d.called
    contenders[3].errback(NameError())
    with pytest.raises(ValueError):
        await d  # first failure is recorded
    assert cancelled == set()


@ensureDeferred
async def test_cancel_early():
    cancelled = set()
    contenders = [
        defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(4)
    ]
    d = transit.there_can_be_only_one(contenders)
    assert not d.called
    assert cancelled == set()
    d.cancel()
    with pytest.raises(defer.CancelledError):
        await d
    assert cancelled == set(range(4))


@ensureDeferred
async def test_cancel_after_one_failure():
    cancelled = set()
    contenders = [
        defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(4)
    ]
    d = transit.there_can_be_only_one(contenders)
    assert not d.called
    assert cancelled == set()
    contenders[0].errback(ValueError())
    d.cancel()
    with pytest.raises(ValueError):
        await d
    assert cancelled == {1, 2, 3}


@pytest.fixture()
def forever():
    clock = task.Clock()
    c = transit.Common("", reactor=clock)
    cancelled = []
    d0 = defer.Deferred(cancelled.append)
    d = c._not_forever(1.0, d0)
    yield c, clock, d0, d, cancelled


@ensureDeferred
async def test_not_forever_fires(forever):
    c, clock, d0, d, cancelled = forever
    assert not d.called
    assert cancelled == []
    d.callback(1)
    answer = await d
    assert answer == 1
    assert cancelled == []
    assert not clock.getDelayedCalls()


@ensureDeferred
async def test_not_forever_errs(forever):
    c, clock, d0, d, cancelled = forever
    assert not d.called
    assert cancelled == []
    d.errback(ValueError())
    assert cancelled == []
    with pytest.raises(ValueError):
        await d
    assert not clock.getDelayedCalls()


@ensureDeferred
async def test_not_forever_cancel_early(forever):
    c, clock, d0, d, cancelled = forever
    assert not d.called
    assert cancelled == []
    d.cancel()
    assert cancelled == [d0]
    with pytest.raises(defer.CancelledError):
        await d
    assert not clock.getDelayedCalls()


@ensureDeferred
async def test_not_forever_timeout(forever):
    c, clock, d0, d, cancelled = forever
    assert not d.called
    assert cancelled == []
    clock.advance(2.0)
    assert cancelled == [d0]
    with pytest.raises(defer.CancelledError):
        await d
    assert not clock.getDelayedCalls()


def test_allocate_port():
    portno = transit.allocate_tcp_port()
    assert isinstance(portno, int)


def test_allocate_port_no_reuseaddr():
    mock_sys = mock.Mock()
    mock_sys.platform = "cygwin"
    with mock.patch("wormhole.transit.sys", mock_sys):
        portno = transit.allocate_tcp_port()
    assert isinstance(portno, int)


LOOPADDR = "127.0.0.1"
OTHERADDR = "1.2.3.4"


@ensureDeferred
async def test_relay_hints():
    URL = "tcp:host:1234"
    c = transit.Common(URL, no_listen=True)
    hints = await c.get_connection_hints()
    assert hints == [{
        "type":
        "relay-v1",
        "hints": [{
            "type": "direct-tcp-v1",
            "hostname": "host",
            "port": 1234,
            "priority": 0.0
        }],
    }]
    with pytest.raises(InternalError):
        transit.Common(123)


@ensureDeferred
async def test_no_relay_hints():
    c = transit.Common(None, no_listen=True)
    hints = await c.get_connection_hints()
    assert hints == []


def test_ignore_bad_hints():
    c = transit.Common("")
    c.add_connection_hints([{"type": "unknown"}])
    c.add_connection_hints([{
        "type": "relay-v1",
        "hints": [{
            "type": "unknown"
        }]
    }])
    assert c._their_direct_hints == []
    assert c._our_relay_hints == set()


@ensureDeferred
async def test_ignore_localhost_hint_orig():
    # this actually starts the listener
    c = transit.TransitSender("")
    hints = await c.get_connection_hints()
    c._stop_listening()
    # If there are non-localhost hints, then localhost hints should be
    # removed. But if the only hint is localhost, it should stay.
    if len(hints) == 1:
        if hints[0]["hostname"] == "127.0.0.1":
            return
    for hint in hints:
        assert not (hint["hostname"] == "127.0.0.1")


@ensureDeferred
async def test_ignore_localhost_hint():
    # this actually starts the listener
    c = transit.TransitSender("")
    with mock.patch(
            "wormhole.ipaddrs.find_addresses",
            return_value=[LOOPADDR, OTHERADDR]):
        hints = await c.get_connection_hints()
    c._stop_listening()
    # If there are non-localhost hints, then localhost hints should be
    # removed.
    assert len(hints) == 1
    assert hints[0]["hostname"] == "1.2.3.4"


@ensureDeferred
async def test_keep_only_localhost_hint():
    # this actually starts the listener
    c = transit.TransitSender("")
    with mock.patch(
            "wormhole.ipaddrs.find_addresses", return_value=[LOOPADDR]):
        hints = await c.get_connection_hints()
    c._stop_listening()
    # If the only hint is localhost, it should stay.
    assert len(hints) == 1
    assert hints[0]["hostname"] == "127.0.0.1"


def test_abilities():
    c = transit.Common(None, no_listen=True)
    abilities = c.get_connection_abilities()
    assert abilities == [
        {
            "type": "direct-tcp-v1"
        },
        {
            "type": "relay-v1"
        },
    ]


@ensureDeferred
async def test_transit_key_wait():
    KEY = b"123"
    c = transit.Common("")
    d = c._get_transit_key()
    assert not d.called
    c.set_transit_key(KEY)
    assert await d == KEY


@ensureDeferred
async def test_transit_key_already_set():
    KEY = b"123"
    c = transit.Common("")
    c.set_transit_key(KEY)
    d = c._get_transit_key()
    assert await d == KEY


async def test_transit_keys():
    KEY = b"123"
    s = transit.TransitSender("")
    s.set_transit_key(KEY)
    r = transit.TransitReceiver("")
    r.set_transit_key(KEY)

    assert s._send_this() == (
        b"transit sender "
        b"559bdeae4b49fa6a23378d2b68f4c7e69378615d4af049c371c6a26e82391089"
        b" ready\n\n")
    assert s._send_this() == r._expect_this()

    assert r._send_this() == (
        b"transit receiver "
        b"ed447528194bac4c00d0c854b12a97ce51413d89aa74d6304475f516fdc23a1b"
        b" ready\n\n")
    assert r._send_this() == s._expect_this()

    assert hexlify(s._sender_record_key()) == \
        b"5a2fba3a9e524ab2e2823ff53b05f946896f6e4ce4e282ffd8e3ac0e5e9e0cda"
    assert hexlify(s._sender_record_key()) == hexlify(r._receiver_record_key())

    assert hexlify(r._sender_record_key()) == \
        b"eedb143117249f45b39da324decf6bd9aae33b7ccd58487436de611a3c6b871d"
    assert hexlify(r._sender_record_key()) == hexlify(s._receiver_record_key())


def test_connection_ready():
    s = transit.TransitSender("")
    assert s.connection_ready("p1") == "go"
    assert s._winner == "p1"
    assert s.connection_ready("p2") == "nevermind"
    assert s._winner == "p1"

    r = transit.TransitReceiver("")
    assert r.connection_ready("p1") == "wait-for-decision"
    assert r.connection_ready("p2") == "wait-for-decision"


def test_listener():
    c = transit.Common("")
    hints, ep = c._build_listener()
    assert isinstance(hints, (list, set))
    if hints:
        assert isinstance(hints[0], DirectTCPV1Hint)
    assert isinstance(ep, endpoints.TCP4ServerEndpoint)



@ensureDeferred
async def test_get_direct_hints():
    # this actually starts the listener
    c = transit.TransitSender("")

    hints = await c.get_connection_hints()

    # the hints are supposed to be cached, so calling this twice won't
    # start a second listener
    assert c._listener
    d2 = c.get_connection_hints()
    assert await d2 == hints

    c._stop_listening()


class DummyProtocol(protocol.Protocol):
    def __init__(self):
        self.buf = b""
        self._count = None
        self._d2 = None

    def wait_for(self, count):
        if len(self.buf) >= count:
            data = self.buf[:count]
            self.buf = self.buf[count:]
            return defer.succeed(data)
        self._d = defer.Deferred()
        self._count = count
        return self._d

    def dataReceived(self, data):
        self.buf += data
        # print("oDR", self._count, len(self.buf))
        if self._count is not None and len(self.buf) >= self._count:
            got = self.buf[:self._count]
            self.buf = self.buf[self._count:]
            self._count = None
            self._d.callback(got)

    def wait_for_disconnect(self):
        self._d2 = defer.Deferred()
        return self._d2

    def connectionLost(self, reason):
        if self._d2:
            self._d2.callback(None)


class FakeTransport:
    signalConnectionLost = True

    def __init__(self, p, peeraddr):
        self.protocol = p
        self._peeraddr = peeraddr
        self._buf = b""
        self._connected = True

    def write(self, data):
        self._buf += data

    def loseConnection(self):
        self._connected = False
        if self.signalConnectionLost:
            self.protocol.connectionLost()

    def getPeer(self):
        return self._peeraddr

    def read_buf(self):
        b = self._buf
        self._buf = b""
        return b


class RandomError(Exception):
    pass


class MockConnection:
    def __init__(self, owner, relay_handshake, start, description):
        self.owner = owner
        self.relay_handshake = relay_handshake
        self.start = start
        self._description = description

        def cancel(d):
            self._cancelled = True

        self._d = defer.Deferred(cancel)
        self._start_negotiation_called = False
        self._cancelled = False

    def startNegotiation(self):
        self._start_negotiation_called = True
        return self._d


def test_describe_inbound():
    f = transit.InboundConnectionFactory(None)
    addrH = address.HostnameAddress("example.com", 1234)
    assert f._describePeer(addrH) == "<-example.com:1234"
    addr4 = address.IPv4Address("TCP", "1.2.3.4", 1234)
    assert f._describePeer(addr4) == "<-1.2.3.4:1234"
    addr6 = address.IPv6Address("TCP", "::1", 1234)
    assert f._describePeer(addr6) == "<-::1:1234"
    addrU = address.UNIXAddress("/dev/unlikely")
    assert f._describePeer(addrU) == "<-UNIXAddress('/dev/unlikely')"


@ensureDeferred
async def test_success_inbound():
    f = transit.InboundConnectionFactory("owner")
    f.protocol = MockConnection
    d = f.whenDone()
    assert not d.called

    addr = address.HostnameAddress("example.com", 1234)
    p = f.buildProtocol(addr)
    assert isinstance(p, MockConnection)
    assert p.owner == "owner"
    assert p.relay_handshake is None
    assert not p._start_negotiation_called
    # meh .start

    # this is normally called from Connection.connectionMade
    f.connectionWasMade(p)
    assert p._start_negotiation_called
    assert not d.called
    assert p._description == "<-example.com:1234"

    p._d.callback(p)
    assert await d == p


@ensureDeferred
async def test_one_fail_one_success():
    f = transit.InboundConnectionFactory("owner")
    f.protocol = MockConnection
    d = f.whenDone()
    assert not d.called

    addr1 = address.HostnameAddress("example.com", 1234)
    addr2 = address.HostnameAddress("example.com", 5678)
    p1 = f.buildProtocol(addr1)
    p2 = f.buildProtocol(addr2)

    f.connectionWasMade(p1)
    f.connectionWasMade(p2)
    assert not d.called

    p1._d.errback(transit.BadHandshake("nope"))
    assert not d.called
    p2._d.callback(p2)
    assert await d == p2


@ensureDeferred
async def test_first_success_wins():
    f = transit.InboundConnectionFactory("owner")
    f.protocol = MockConnection
    d = f.whenDone()
    assert not d.called

    addr1 = address.HostnameAddress("example.com", 1234)
    addr2 = address.HostnameAddress("example.com", 5678)
    p1 = f.buildProtocol(addr1)
    p2 = f.buildProtocol(addr2)

    f.connectionWasMade(p1)
    f.connectionWasMade(p2)
    assert not d.called

    p1._d.callback(p1)
    assert await d == p1
    assert not p1._cancelled
    assert p2._cancelled


@ensureDeferred
async def test_cancel():
    f = transit.InboundConnectionFactory("owner")
    f.protocol = MockConnection
    d = f.whenDone()
    assert not d.called

    addr1 = address.HostnameAddress("example.com", 1234)
    addr2 = address.HostnameAddress("example.com", 5678)
    p1 = f.buildProtocol(addr1)
    p2 = f.buildProtocol(addr2)

    f.connectionWasMade(p1)
    f.connectionWasMade(p2)
    assert not d.called

    d.cancel()

    with pytest.raises(defer.CancelledError):
        await d
    assert p1._cancelled
    assert p2._cancelled


def test_success_outboun():
    f = transit.OutboundConnectionFactory("owner", "relay_handshake",
                                          "description")
    f.protocol = MockConnection

    addr = address.HostnameAddress("example.com", 1234)
    p = f.buildProtocol(addr)
    assert isinstance(p, MockConnection)
    assert p.owner == "owner"
    assert p.relay_handshake == "relay_handshake"
    assert not p._start_negotiation_called
    # meh .start

    # this is normally called from Connection.connectionMade
    f.connectionWasMade(p)  # no-op for outbound
    assert not p._start_negotiation_called


class MockOwner:
    _connection_ready_called = False

    def connection_ready(self, connection):
        self._connection_ready_called = True
        self._connection = connection
        return self._state

    def _send_this(self):
        return b"send_this"

    def _expect_this(self):
        return b"expect_this"

    def _sender_record_key(self):
        return b"s" * 32

    def _receiver_record_key(self):
        return b"r" * 32


class MockFactory:
    _connectionWasMade_called = False

    def connectionWasMade(self, p):
        self._connectionWasMade_called = True
        self._p = p


def test_check_and_remove():
    c = transit.Connection(None, None, None, "description")
    c.buf = b""
    EXP = b"expectation"
    assert not c._check_and_remove(EXP)
    assert c.buf == b""

    c.buf = b"unexpected"
    with pytest.raises(transit.BadHandshake) as f:
        c._check_and_remove(EXP)
    assert str(f.value) == f"got {b'unexpected'!r} want {b'expectation'!r}"
    assert c.buf == b"unexpected"

    c.buf = b"expect"
    assert not c._check_and_remove(EXP)
    assert c.buf == b"expect"

    c.buf = b"expectation"
    assert c._check_and_remove(EXP)
    assert c.buf == b""

    c.buf = b"expectation exceeded"
    assert c._check_and_remove(EXP)
    assert c.buf == b" exceeded"


def test_describe_connection():
    c = transit.Connection(None, None, None, "description")
    assert c.describe() == "description"


@ensureDeferred
async def test_sender_accepting():
    relay_handshake = None
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, relay_handshake, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    owner._state = "go"
    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called

    c.dataReceived(b"expect_this")
    assert t.read_buf() == b"go\n"
    assert t._connected
    assert c.state == "records"
    assert await d == c

    c.close()
    assert not t._connected


@ensureDeferred
async def test_sender_rejecting():
    relay_handshake = None
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, relay_handshake, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    owner._state = "nevermind"
    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called

    c.dataReceived(b"expect_this")
    assert t.read_buf() == b"nevermind\n"
    assert not t._connected
    assert c.state == "hung up"
    with pytest.raises(transit.BadHandshake) as f:
        await d
    assert str(f.value) == "abandoned"


@ensureDeferred
async def test_handshake_other_error():
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called
    c.state = RandomError("boom2")
    with pytest.raises(RandomError):
        c.dataReceived(b"surprise!")
    assert not t._connected
    assert c.state == "hung up"
    with pytest.raises(RandomError):
        await d


@ensureDeferred
async def test_handshake_bad_state():
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called
    c.state = "unknown-bogus-state"
    with pytest.raises(ValueError):
        c.dataReceived(b"surprise!")
    assert not t._connected
    assert c.state == "hung up"
    with pytest.raises(ValueError):
        await d


@ensureDeferred
async def test_relay_handshake():
    relay_handshake = b"relay handshake"
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, relay_handshake, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c
    assert t.read_buf() == b""  # quiet until startNegotiation

    owner._state = "go"
    d = c.startNegotiation()
    assert t.read_buf() == relay_handshake
    assert c.state == "relay"  # waiting for OK from relay

    c.dataReceived(b"ok\n")
    assert t.read_buf() == b"send_this"
    assert c.state == "handshake"

    assert not d.called

    c.dataReceived(b"expect_this")
    assert c.state == "records"
    assert await d == c

    assert t.read_buf() == b"go\n"


@ensureDeferred
async def test_relay_handshake_bad():
    relay_handshake = b"relay handshake"
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, relay_handshake, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c
    assert t.read_buf() == b""  # quiet until startNegotiation

    owner._state = "go"
    d = c.startNegotiation()
    assert t.read_buf() == relay_handshake
    assert c.state == "relay"  # waiting for OK from relay

    c.dataReceived(b"not ok\n")
    assert not t._connected
    assert c.state == "hung up"

    with pytest.raises(transit.BadHandshake) as f:
        await d
    assert str(f.value) == "got {!r} want {!r}".format(b"not ok\n", b"ok\n")


@ensureDeferred
async def test_receiver_accepted():
    # we're on the receiving side, so we wait for the sender to decide
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    owner._state = "wait-for-decision"
    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called

    c.dataReceived(b"expect_this")
    assert c.state == "wait-for-decision"
    assert not d.called

    c.dataReceived(b"go\n")
    assert c.state == "records"
    assert await d == c


@ensureDeferred
async def test_receiver_rejected_politely():
    # we're on the receiving side, so we wait for the sender to decide
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    owner._state = "wait-for-decision"
    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called

    c.dataReceived(b"expect_this")
    assert c.state == "wait-for-decision"
    assert not d.called

    c.dataReceived(b"nevermind\n")  # polite rejection
    assert not t._connected
    assert c.state == "hung up"
    with pytest.raises(transit.BadHandshake) as f:
        await d
    assert str(f.value) == "got {!r} want {!r}".format(b"nevermind\n", b"go\n")


@ensureDeferred
async def test_receiver_rejected_rudely():
    # we're on the receiving side, so we wait for the sender to decide
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    assert factory._connectionWasMade_called
    assert factory._p == c

    owner._state = "wait-for-decision"
    d = c.startNegotiation()
    assert c.state == "handshake"
    assert t.read_buf() == b"send_this"
    assert not d.called

    c.dataReceived(b"expect_this")
    assert c.state == "wait-for-decision"
    assert not d.called

    t.loseConnection()
    assert not t._connected
    with pytest.raises(transit.BadHandshake) as f:
        await d
    assert str(f.value) == "connection lost"


@ensureDeferred
async def test_cancel_during_negotiation():
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()

    d = c.startNegotiation()
    # while we're waiting for negotiation, we get cancelled
    d.cancel()

    assert not t._connected
    assert c.state == "hung up"
    with pytest.raises(defer.CancelledError):
        await d


@ensureDeferred
async def test_timeout():
    clock = task.Clock()
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")

    def _callLater(period, func):
        clock.callLater(period, func)

    c.callLater = _callLater
    assert c.state == "too-early"
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()
    # the timer should now be running
    d = c.startNegotiation()
    # while we're waiting for negotiation, the timer expires
    clock.advance(transit.TIMEOUT + 1.0)

    assert not t._connected
    with pytest.raises(transit.BadHandshake) as f:
        await d
    assert str(f.value) == "timeout"


async def make_connection():
    owner = MockOwner()
    factory = MockFactory()
    addr = address.HostnameAddress("example.com", 1234)
    c = transit.Connection(owner, None, None, "description")
    t = c.transport = FakeTransport(c, addr)
    c.factory = factory
    c.connectionMade()

    owner._state = "go"
    d = c.startNegotiation()
    c.dataReceived(b"expect_this")
    assert await d == c
    t.read_buf()  # flush input buffer, prepare for encrypted records

    return t, c, owner


@ensureDeferred
async def test_records_not_binary():
    t, c, owner = await make_connection()

    RECORD1 = "not binary"
    with pytest.raises(InternalError):
        c.send_record(RECORD1)


@ensureDeferred
async def test_records_good():
    # now make sure that outbound records are encrypted properly
    t, c, owner = await make_connection()

    RECORD1 = b"record"
    c.send_record(RECORD1)
    buf = t.read_buf()
    expected = f"{24 + len(RECORD1) + 16:08x}".encode("ascii")
    assert hexlify(buf[:4]) == expected
    encrypted = buf[4:]
    receive_box = SecretBox(owner._sender_record_key())
    nonce_buf = encrypted[:SecretBox.NONCE_SIZE]  # assume it's prepended
    nonce = int(hexlify(nonce_buf), 16)
    assert nonce == 0  # first message gets nonce 0
    decrypted = receive_box.decrypt(encrypted)
    assert decrypted == RECORD1

    # second message gets nonce 1
    RECORD2 = b"record2"
    c.send_record(RECORD2)
    buf = t.read_buf()
    expected = f"{24 + len(RECORD2) + 16:08x}".encode("ascii")
    assert hexlify(buf[:4]) == expected
    encrypted = buf[4:]
    receive_box = SecretBox(owner._sender_record_key())
    nonce_buf = encrypted[:SecretBox.NONCE_SIZE]  # assume it's prepended
    nonce = int(hexlify(nonce_buf), 16)
    assert nonce == 1
    decrypted = receive_box.decrypt(encrypted)
    assert decrypted == RECORD2

    # and that we can receive records properly
    inbound_records = []
    c.recordReceived = inbound_records.append
    send_box = SecretBox(owner._receiver_record_key())

    RECORD3 = b"record3"
    nonce_buf = unhexlify("%048x" % 0)  # first nonce must be 0
    encrypted = send_box.encrypt(RECORD3, nonce_buf)
    length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
    c.dataReceived(length[:2])
    c.dataReceived(length[2:])
    c.dataReceived(encrypted[:-2])
    assert inbound_records == []
    c.dataReceived(encrypted[-2:])
    assert inbound_records == [RECORD3]

    RECORD4 = b"record4"
    nonce_buf = unhexlify("%048x" % 1)  # nonces increment
    encrypted = send_box.encrypt(RECORD4, nonce_buf)
    length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
    c.dataReceived(length[:2])
    c.dataReceived(length[2:])
    c.dataReceived(encrypted[:-2])
    assert inbound_records == [RECORD3]
    c.dataReceived(encrypted[-2:])
    assert inbound_records == [RECORD3, RECORD4]

    # receiving two records at the same time: deliver both
    inbound_records[:] = []
    RECORD5 = b"record5"
    nonce_buf = unhexlify("%048x" % 2)  # nonces increment
    encrypted = send_box.encrypt(RECORD5, nonce_buf)
    length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
    r5 = length + encrypted
    RECORD6 = b"record6"
    nonce_buf = unhexlify("%048x" % 3)  # nonces increment
    encrypted = send_box.encrypt(RECORD6, nonce_buf)
    length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
    r6 = length + encrypted
    c.dataReceived(r5 + r6)
    assert inbound_records == [RECORD5, RECORD6]


def corrupt(orig):
    last_byte = orig[-1:]
    num = int(hexlify(last_byte).decode("ascii"), 16)
    corrupt_num = 256 - num
    as_byte = unhexlify(f"{corrupt_num:02x}")
    return orig[:-1] + as_byte


@ensureDeferred
async def test_records_corrupt():
    # corrupt records should be rejected
    t, c, owner = await make_connection()

    inbound_records = []
    c.recordReceived = inbound_records.append

    RECORD = b"record"
    send_box = SecretBox(owner._receiver_record_key())
    nonce_buf = unhexlify("%048x" % 0)  # first nonce must be 0
    encrypted = corrupt(send_box.encrypt(RECORD, nonce_buf))
    length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
    c.dataReceived(length)
    c.dataReceived(encrypted[:-2])
    assert inbound_records == []
    with pytest.raises(CryptoError):
        c.dataReceived(encrypted[-2:])
    assert inbound_records == []
    # and the connection should have been dropped
    assert not t._connected


@ensureDeferred
async def test_out_of_order_nonce():
    # an inbound out-of-order nonce should be rejected
    t, c, owner = await make_connection()

    inbound_records = []
    c.recordReceived = inbound_records.append

    RECORD = b"record"
    send_box = SecretBox(owner._receiver_record_key())
    nonce_buf = unhexlify("%048x" % 1)  # first nonce must be 0
    encrypted = send_box.encrypt(RECORD, nonce_buf)
    length = unhexlify(f"{len(encrypted):08x}")  # always 4 bytes long
    c.dataReceived(length)
    c.dataReceived(encrypted[:-2])
    assert inbound_records == []
    with pytest.raises(transit.BadNonce):
        c.dataReceived(encrypted[-2:])
    assert inbound_records == []
    # and the connection should have been dropped
    assert not t._connected


# TODO: check that .connectionLost/loseConnection signatures are
# consistent: zero args, or one arg?

# XXX: if we don't set the transit key before connecting, what
# happens? We currently get a type-check assertion from HKDF because
# the key is None.


@ensureDeferred
async def test_receive_queue():
    c = transit.Connection(None, None, None, "description")
    c.transport = FakeTransport(c, None)
    c.transport.signalConnectionLost = False
    c.recordReceived(b"0")
    c.recordReceived(b"1")
    c.recordReceived(b"2")
    d0 = c.receive_record()
    assert await d0 == b"0"
    d1 = c.receive_record()
    d2 = c.receive_record()
    # they must fire in order of receipt, not order of addCallback
    assert await d2 == b"2"
    assert await d1 == b"1"

    d3 = c.receive_record()
    d4 = c.receive_record()
    assert not d3.called
    assert not d4.called

    c.recordReceived(b"3")
    assert await d3 == b"3"
    assert not d4.called

    c.recordReceived(b"4")
    assert await d4 == b"4"

    d5 = c.receive_record()
    c.close()
    with pytest.raises(error.ConnectionClosed):
        await d5


@ensureDeferred
async def test_receive_queue_connectionLost():
    """
    Test that pending receive_record() calls are errback'd when the
    connection is lost from the remote side
    """
    c = transit.Connection(None, None, None, "description")
    c.transport = FakeTransport(c, None)
    c.transport.signalConnectionLost = False
    c._negotiation_d.addErrback(lambda err: None)  # eat negotiation error

    d1 = c.receive_record()
    d2 = c.receive_record()
    assert not d1.called
    assert not d2.called

    c.connectionLost()

    with pytest.raises(error.ConnectionClosed):
        await d1
    with pytest.raises(error.ConnectionClosed):
        await d2


def test_producer():
    # a Transit object (receiving data from the remote peer) produces
    # data and writes it into a local Consumer
    c = transit.Connection(None, None, None, "description")
    c.transport = proto_helpers.StringTransport()
    c.recordReceived(b"r1.")
    c.recordReceived(b"r2.")

    consumer = proto_helpers.StringTransport()
    rv = c.connectConsumer(consumer)
    assert rv is None
    assert c._consumer is consumer
    assert consumer.value() == b"r1.r2."

    with pytest.raises(RuntimeError):
        c.connectConsumer(consumer)

    c.recordReceived(b"r3.")
    assert consumer.value() == b"r1.r2.r3."

    c.pauseProducing()
    assert c.transport.producerState == "paused"
    c.resumeProducing()
    assert c.transport.producerState == "producing"

    c.disconnectConsumer()
    assert consumer.producer is None
    c.connectConsumer(consumer)

    c.stopProducing()
    assert c.transport.producerState == "stopped"


@ensureDeferred
async def test_connectConsumer():
    # connectConsumer() takes an optional number of bytes to expect, and
    # fires a Deferred when that many have been written
    c = transit.Connection(None, None, None, "description")
    c._negotiation_d.addErrback(lambda err: None)  # eat it
    c.transport = proto_helpers.StringTransport()
    c.recordReceived(b"r1.")

    consumer = proto_helpers.StringTransport()
    d = c.connectConsumer(consumer, expected=10)
    assert consumer.value() == b"r1."
    assert not d.called

    c.recordReceived(b"r2.")
    assert consumer.value() == b"r1.r2."
    assert not d.called

    c.recordReceived(b"r3.")
    assert consumer.value() == b"r1.r2.r3."
    assert not d.called

    c.recordReceived(b"!")
    assert consumer.value() == b"r1.r2.r3.!"
    assert await d == 10

    # that should automatically disconnect the consumer, and subsequent
    # records should get queued, not delivered
    assert c._consumer is None
    c.recordReceived(b"overflow")
    assert consumer.value() == b"r1.r2.r3.!"

    # now test that the Deferred errbacks when the connection is lost
    d = c.connectConsumer(consumer, expected=10)

    c.connectionLost()
    with pytest.raises(error.ConnectionClosed):
        await d


@ensureDeferred
async def test_connectConsumer_empty():
    # if connectConsumer() expects 0 bytes (e.g. someone is "sending" a
    # zero-length file), make sure it gets woken up right away, so it can
    # disconnect itself, even though no bytes will actually arrive
    c = transit.Connection(None, None, None, "description")
    c._negotiation_d.addErrback(lambda err: None)  # eat it
    c.transport = proto_helpers.StringTransport()

    consumer = proto_helpers.StringTransport()
    d = c.connectConsumer(consumer, expected=0)
    assert await d == 0
    assert consumer.value() == b""
    assert c._consumer is None


@ensureDeferred
async def test_writeToFile():
    c = transit.Connection(None, None, None, "description")
    c._negotiation_d.addErrback(lambda err: None)  # eat it
    c.transport = proto_helpers.StringTransport()
    c.recordReceived(b"r1.")

    f = io.BytesIO()
    progress = []
    d = c.writeToFile(f, 10, progress.append)
    assert f.getvalue() == b"r1."
    assert progress == [3]
    assert not d.called

    c.recordReceived(b"r2.")
    assert f.getvalue() == b"r1.r2."
    assert progress == [3, 3]
    assert not d.called

    c.recordReceived(b"r3.")
    assert f.getvalue() == b"r1.r2.r3."
    assert progress == [3, 3, 3]
    assert not d.called

    c.recordReceived(b"!")
    assert f.getvalue() == b"r1.r2.r3.!"
    assert progress == [3, 3, 3, 1]
    assert await d == 10

    # that should automatically disconnect the consumer, and subsequent
    # records should get queued, not delivered
    assert c._consumer is None
    c.recordReceived(b"overflow.")
    assert f.getvalue() == b"r1.r2.r3.!"
    assert progress == [3, 3, 3, 1]

    # test what happens when enough data is queued ahead of time
    c.recordReceived(b"second.")  # now "overflow.second."
    c.recordReceived(b"third.")  # now "overflow.second.third."
    f = io.BytesIO()
    d = c.writeToFile(f, 10)
    assert f.getvalue() == b"overflow.second."  # whole records
    assert await d == 16
    assert list(c._inbound_records) == [b"third."]

    # now test that the Deferred errbacks when the connection is lost
    d = c.writeToFile(f, 10)

    c.connectionLost()
    with pytest.raises(error.ConnectionClosed):
        await d


def test_consumer():
    # a local producer sends data to a consuming Transit object
    c = transit.Connection(None, None, None, "description")
    c.transport = proto_helpers.StringTransport()
    records = []
    c.send_record = records.append

    producer = proto_helpers.StringTransport()
    c.registerProducer(producer, True)
    assert c.transport.producer is producer

    c.write(b"r1.")
    assert records == [b"r1."]

    c.unregisterProducer()
    assert c.transport.producer is None


def test_basic():
    f = io.BytesIO()
    progress = []
    fc = transit.FileConsumer(f, progress.append)
    assert progress == []
    assert f.getvalue() == b""
    fc.write(b"." * 99)
    assert progress == [99]
    assert f.getvalue() == b"." * 99
    fc.write(b"!")
    assert progress == [99, 1]
    assert f.getvalue() == b"." * 99 + b"!"


def test_hasher():
    hashee = []
    f = io.BytesIO()
    progress = []
    fc = transit.FileConsumer(f, progress.append, hasher=hashee.append)
    assert progress == []
    assert f.getvalue() == b""
    assert hashee == []
    fc.write(b"." * 99)
    assert progress == [99]
    assert f.getvalue() == b"." * 99
    assert hashee == [b"." * 99]
    fc.write(b"!")
    assert progress == [99, 1]
    assert f.getvalue() == b"." * 99 + b"!"
    assert hashee == [b"." * 99, b"!"]


DIRECT_HINT_JSON = {
    "type": "direct-tcp-v1",
    "hostname": "direct",
    "port": 1234
}
RELAY_HINT_JSON = {
    "type": "relay-v1",
    "hints": [{
        "type": "direct-tcp-v1",
        "hostname": "relay",
        "port": 1234
    }]
}
UNRECOGNIZED_DIRECT_HINT_JSON = {
    "type": "direct-tcp-v1",
    "hostname": ["cannot", "parse", "list"]
}
UNRECOGNIZED_HINT_JSON = {"type": "unknown"}
UNAVAILABLE_HINT_JSON = {
    "type": "direct-tcp-v1",  # e.g. Tor without txtorcon
    "hostname": "unavailable",
    "port": 1234
}
RELAY_HINT2_JSON = {
    "type":
    "relay-v1",
    "hints": [{
        "type": "direct-tcp-v1",
        "hostname": "relay",
        "port": 1234
    }, UNRECOGNIZED_HINT_JSON]
}
UNAVAILABLE_RELAY_HINT_JSON = {
    "type": "relay-v1",
    "hints": [UNAVAILABLE_HINT_JSON]
}


class FakeConnector:
    def __init__(self):
        self._connectors = []
        self._waiters = []
        self._descriptions = []

    def _start_connector(self, ep, description, is_relay=False):
        d = defer.Deferred()
        self._connectors.append(ep)
        self._waiters.append(d)
        self._descriptions.append(description)
        return d


@ensureDeferred
async def test_success_direct():
    reactor = mock.Mock()
    fc = FakeConnector()
    s = transit.TransitSender("", reactor=reactor)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()  # start the listener
    del hints
    s.add_connection_hints([
        DIRECT_HINT_JSON, UNRECOGNIZED_DIRECT_HINT_JSON,
        UNRECOGNIZED_HINT_JSON
    ])

    s._start_connector = fc._start_connector
    d = s.connect()
    assert not d.called
    assert len(fc._waiters) == 1
    assert isinstance(fc._waiters[0], defer.Deferred)

    fc._waiters[0].callback("winner")
    assert await d == "winner"
    assert fc._descriptions == ["->tcp:direct:1234"]


@ensureDeferred
async def test_success_direct_tor():
    clock = task.Clock()
    fc = FakeConnector()
    s = transit.TransitSender("", tor=mock.Mock(), reactor=clock)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()  # start the listener
    del hints
    s.add_connection_hints([DIRECT_HINT_JSON])

    s._start_connector = fc._start_connector
    d = s.connect()
    assert not d.called
    assert len(fc._waiters) == 1
    assert isinstance(fc._waiters[0], defer.Deferred)

    fc._waiters[0].callback("winner")
    assert await d == "winner"
    assert fc._descriptions == ["tor->tcp:direct:1234"]


@ensureDeferred
async def test_success_direct_tor_relay():
    clock = task.Clock()
    fc = FakeConnector()
    s = transit.TransitSender("", tor=mock.Mock(), reactor=clock)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()  # start the listener
    del hints
    s.add_connection_hints([RELAY_HINT_JSON])

    s._start_connector = fc._start_connector
    d = s.connect()
    # move the clock forward any amount, since relay connections are
    # triggered starting at T+0.0
    clock.advance(1.0)
    assert not d.called
    assert len(fc._waiters) == 1
    assert isinstance(fc._waiters[0], defer.Deferred)

    fc._waiters[0].callback("winner")
    assert await d == "winner"
    assert fc._descriptions == ["tor->relay:tcp:relay:1234"]


def _endpoint_from_hint_obj(hint, _tor, _reactor):
    if isinstance(hint, DirectTCPV1Hint):
        if hint.hostname == "unavailable":
            return None
        return hint.hostname
    return None


@ensureDeferred
async def test_wait_for_relay():
    clock = task.Clock()
    fc = FakeConnector()
    s = transit.TransitSender("", reactor=clock, no_listen=True)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()
    del hints
    s.add_connection_hints(
        [DIRECT_HINT_JSON, UNRECOGNIZED_HINT_JSON, RELAY_HINT_JSON])
    s._start_connector = fc._start_connector

    with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                    _endpoint_from_hint_obj):
        d = s.connect()
        assert not d.called
        # the direct connectors are tried right away, but the relay
        # connectors are stalled for a few seconds
        assert fc._connectors == ["direct"]

        clock.advance(s.RELAY_DELAY + 1.0)
        assert fc._connectors == ["direct", "relay"]

        fc._waiters[0].callback("winner")
        assert await d == "winner"


@ensureDeferred
async def test_priorities():
    clock = task.Clock()
    fc = FakeConnector()
    s = transit.TransitSender("", reactor=clock, no_listen=True)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()
    del hints
    s.add_connection_hints([
        {
            "type":
            "relay-v1",
            "hints": [{
                "type": "direct-tcp-v1",
                "hostname": "relay",
                "port": 1234
            }]
        },
        {
            "type": "direct-tcp-v1",
            "hostname": "direct",
            "port": 1234
        },
        {
            "type":
            "relay-v1",
            "hints": [{
                "type": "direct-tcp-v1",
                "priority": 2.0,
                "hostname": "relay2",
                "port": 1234
            }, {
                "type": "direct-tcp-v1",
                "priority": 3.0,
                "hostname": "relay3",
                "port": 1234
            }]
        },
        {
            "type":
            "relay-v1",
            "hints": [{
                "type": "direct-tcp-v1",
                "priority": 2.0,
                "hostname": "relay4",
                "port": 1234
            }]
        },
    ])
    s._start_connector = fc._start_connector

    with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                    _endpoint_from_hint_obj):
        d = s.connect()
        assert not d.called
        # direct connector should be used first, then the priority=3.0 relay,
        # then the two 2.0 relays, then the (default) 0.0 relay

        assert fc._connectors == ["direct"]

        clock.advance(s.RELAY_DELAY + 1.0)
        assert fc._connectors == ["direct", "relay3"]

        clock.advance(s.RELAY_DELAY)
        assert fc._connectors in \
                      (["direct", "relay3", "relay2", "relay4"],
                       ["direct", "relay3", "relay4", "relay2"])

        clock.advance(s.RELAY_DELAY)
        assert fc._connectors in \
                      (["direct", "relay3", "relay2", "relay4", "relay"],
                       ["direct", "relay3", "relay4", "relay2", "relay"])

        fc._waiters[0].callback("winner")
        assert await d == "winner"


@ensureDeferred
async def test_no_direct_hints():
    clock = task.Clock()
    fc = FakeConnector()
    s = transit.TransitSender("", reactor=clock, no_listen=True)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()  # start the listener
    del hints
    # include hints that can't be turned into an endpoint at runtime
    s.add_connection_hints([
        UNRECOGNIZED_HINT_JSON, UNAVAILABLE_HINT_JSON, RELAY_HINT2_JSON,
        UNAVAILABLE_RELAY_HINT_JSON
    ])
    s._start_connector = fc._start_connector

    with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                    _endpoint_from_hint_obj):
        d = s.connect()
        assert not d.called
        # since there are no usable direct hints, the relay connector will
        # only be stalled for 0 seconds
        assert fc._connectors == []

        clock.advance(0)
        assert fc._connectors == ["relay"]

        fc._waiters[0].callback("winner")
        assert await d == "winner"


@ensureDeferred
async def test_no_contenders():
    clock = task.Clock()
    fc = FakeConnector()
    s = transit.TransitSender("", reactor=clock, no_listen=True)
    s.set_transit_key(b"key")
    hints = await s.get_connection_hints()  # start the listener
    del hints
    s.add_connection_hints([])  # no hints at all
    s._start_connector = fc._start_connector

    with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                    _endpoint_from_hint_obj):
        d = s.connect()
        with pytest.raises(transit.TransitError) as f:
            await d
        assert str(f.value) == "No contenders for connection"


def old_build_relay_handshake(key):
    token = HKDF(key, 32, CTXinfo=b"transit_relay_token")
    return (token, b"please relay " + hexlify(token) + b"\n")



def test_old():
    key = b"\x00"
    token, old_handshake = old_build_relay_handshake(key)
    tc = transit_server.TransitConnection()
    tc.factory = mock.Mock()
    tc.factory.connection_got_token = mock.Mock()
    tc.transport = mock.Mock()
    tc.connectionMade()
    tc._state.please_relay = mock.Mock()
    tc._state.please_relay_for_side = mock.Mock()
    tc.dataReceived(old_handshake[:-1])
    assert tc._state.please_relay.mock_calls == []
    tc.dataReceived(old_handshake[-1:])
    assert tc._state.please_relay.mock_calls == \
                     [mock.call(hexlify(token))]
    assert tc._state.please_relay_for_side.mock_calls == []


def test_new():
    c = transit.Common(None)
    c.set_transit_key(b"\x00")
    new_handshake = c._build_relay_handshake()
    token, old_handshake = old_build_relay_handshake(b"\x00")

    tc = transit_server.TransitConnection()
    tc.factory = mock.Mock()
    tc.transport = mock.Mock()
    tc.connectionMade()
    tc._state._real_register_token_for_side = m = mock.Mock()
    tc.dataReceived(new_handshake[:-1])
    assert m.mock_calls == []
    tc.dataReceived(new_handshake[-1:])
    assert m.mock_calls == \
        [mock.call(hexlify(token), c._side.encode("ascii"))]


async def doBoth(d1, d2):
    return await gatherResults([d1, d2], True)


@ensureDeferred
async def test_direct():
    KEY = b"k" * 32
    s = transit.TransitSender(None)
    r = transit.TransitReceiver(None)

    s.set_transit_key(KEY)
    r.set_transit_key(KEY)

    # TODO: this sometimes fails with EADDRINUSE
    shints = await s.get_connection_hints()
    rhints = await r.get_connection_hints()

    s.add_connection_hints(rhints)
    r.add_connection_hints(shints)

    (x, y) = await doBoth(s.connect(), r.connect())
    assert isinstance(x, transit.Connection)
    assert isinstance(y, transit.Connection)

    d = y.receive_record()

    x.send_record(b"record1")
    r = await d
    assert r == b"record1"

    x.close()
    y.close()


@ensureDeferred
async def test_relay(transit_relay):
    KEY = b"k" * 32
    s = transit.TransitSender(transit_relay, no_listen=True)
    r = transit.TransitReceiver(transit_relay, no_listen=True)

    s.set_transit_key(KEY)
    r.set_transit_key(KEY)

    shints = await s.get_connection_hints()
    rhints = await r.get_connection_hints()

    s.add_connection_hints(rhints)
    r.add_connection_hints(shints)

    (x, y) = await doBoth(s.connect(), r.connect())
    assert isinstance(x, transit.Connection)
    assert isinstance(y, transit.Connection)

    d = y.receive_record()

    x.send_record(b"record1")
    r = await d
    assert r == b"record1"

    x.close()
    y.close()



from attr import attrib, attrs
from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces
from .util import provides


@attrs
@implementer(_interfaces.ILister)
class Lister:
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def wire(self, rendezvous_connector, input):
        self._RC = _interfaces.IRendezvousConnector(rendezvous_connector)
        self._I = _interfaces.IInput(input)

    # Ideally, each API request would spawn a new "list_nameplates" message
    # to the server, so the response would be maximally fresh, but that would
    # require correlating server request+response messages, and the protocol
    # is intended to be less stateful than that. So we offer a weaker
    # freshness property: if no server requests are in flight, then a new API
    # request will provoke a new server request, and the result will be
    # fresh. But if a server request is already in flight when a second API
    # request arrives, both requests will be satisfied by the same response.

    @m.state(initial=True)
    def S0A_idle_disconnected(self):
        pass  # pragma: no cover

    @m.state()
    def S1A_wanting_disconnected(self):
        pass  # pragma: no cover

    @m.state()
    def S0B_idle_connected(self):
        pass  # pragma: no cover

    @m.state()
    def S1B_wanting_connected(self):
        pass  # pragma: no cover

    @m.input()
    def connected(self):
        pass

    @m.input()
    def lost(self):
        pass

    @m.input()
    def refresh(self):
        pass

    @m.input()
    def rx_nameplates(self, all_nameplates):
        pass

    @m.output()
    def RC_tx_list(self):
        self._RC.tx_list()

    @m.output()
    def I_got_nameplates(self, all_nameplates):
        # We get a set of nameplate ids. There may be more attributes in the
        # future: change RendezvousConnector._response_handle_nameplates to
        # get them
        self._I.got_nameplates(all_nameplates)

    S0A_idle_disconnected.upon(connected, enter=S0B_idle_connected, outputs=[])
    S0B_idle_connected.upon(lost, enter=S0A_idle_disconnected, outputs=[])

    S0A_idle_disconnected.upon(
        refresh, enter=S1A_wanting_disconnected, outputs=[])
    S1A_wanting_disconnected.upon(
        refresh, enter=S1A_wanting_disconnected, outputs=[])
    S1A_wanting_disconnected.upon(
        connected, enter=S1B_wanting_connected, outputs=[RC_tx_list])
    S0B_idle_connected.upon(
        refresh, enter=S1B_wanting_connected, outputs=[RC_tx_list])
    S0B_idle_connected.upon(
        rx_nameplates, enter=S0B_idle_connected, outputs=[I_got_nameplates])
    S1B_wanting_connected.upon(
        lost, enter=S1A_wanting_disconnected, outputs=[])
    S1B_wanting_connected.upon(
        refresh, enter=S1B_wanting_connected, outputs=[RC_tx_list])
    S1B_wanting_connected.upon(
        rx_nameplates, enter=S0B_idle_connected, outputs=[I_got_nameplates])



from binascii import unhexlify  # , hexlify

from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives import hashes

from .. import util

# def generate_KAT():
#     print("KAT = [")
#     for salt in (b"", b"salt"):
#         for context in (b"", b"context"):
#             skm = b"secret"
#             out = HKDF(skm, 64, XTS=salt, CTXinfo=context)
#             hexout = "  '%s' +\n  '%s'" % (hexlify(out[:32]),
#                                            hexlify(out[32:]))
#             print(" (%r, %r, %r,\n%s)," % (salt, context, skm, hexout))
#     print("]")

KAT = [
    ('', '', 'secret',
     '2f34e5ff91ec85d53ca9b543683174d0cf550b60d5f52b24c97b386cfcf6cbbf' +
     '9cfd42fd37e1e5a214d15f03058d7fee63dc28f564b7b9fe3da514f80daad4bf'),
    ('', 'context', 'secret',
     'c24c303a1adfb4c3e2b092e6254ed481c41d8955ba8ec3f6a1473493a60c957b' +
     '31b723018ca75557214d3d5c61c0c7a5315b103b21ff00cb03ebe023dc347a47'),
    ('salt', '', 'secret',
     'f1156507c39b0e326159e778696253122de430899a8df2484040a85a5f95ceb1' +
     'dfca555d4cc603bdf7153ed1560de8cbc3234b27a6d2be8e8ca202d90649679a'),
    ('salt', 'context', 'secret',
     '61a4f201a867bcc12381ddb180d27074408d03ee9d5750855e5a12d967fa060f' +
     '10336ead9370927eaabb0d60b259346ee5f57eb7ceba8c72f1ed3f2932b1bf19'),
]


def test_kat():
    for (salt, context, skm, expected_hexout) in KAT:
        expected_out = unhexlify(expected_hexout)
        for outlen in range(0, len(expected_out)):
            out = HKDF(
                algorithm=hashes.SHA256(),
                length=outlen,
                salt=salt.encode("ascii"),
                info=context.encode("ascii"),
            ).derive(skm.encode("ascii"))
            assert out == expected_out[:outlen]

def test_util():
    for (salt, context, skm, expected_hexout) in KAT:
        expected_out = unhexlify(expected_hexout)
        for outlen in range(0, len(expected_out)):
            out = util.HKDF(
                skm.encode("ascii"),
                outlen,
                salt.encode("ascii"),
                context.encode("ascii"),
            )
            assert out == expected_out[:outlen]


# if __name__ == '__main__':
#     generate_KAT()



from zope.interface import implementer
from attr import attrs, attrib
from automat import MethodicalMachine
from . import _interfaces
from ._nameplate import validate_nameplate
from .errors import KeyFormatError
from .util import provides


def validate_code(code):
    if ' ' in code:
        raise KeyFormatError(f"Code '{code}' contains spaces.")
    nameplate = code.split("-", 2)[0]
    validate_nameplate(nameplate)  # can raise KeyFormatError


def first(outputs):
    return list(outputs)[0]


@attrs
@implementer(_interfaces.ICode)
class Code:
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def wire(self, boss, allocator, nameplate, key, input):
        self._B = _interfaces.IBoss(boss)
        self._A = _interfaces.IAllocator(allocator)
        self._N = _interfaces.INameplate(nameplate)
        self._K = _interfaces.IKey(key)
        self._I = _interfaces.IInput(input)

    @m.state(initial=True)
    def S0_idle(self):
        pass  # pragma: no cover

    @m.state()
    def S1_inputting_nameplate(self):
        pass  # pragma: no cover

    @m.state()
    def S2_inputting_words(self):
        pass  # pragma: no cover

    @m.state()
    def S3_allocating(self):
        pass  # pragma: no cover

    @m.state()
    def S4_known(self):
        pass  # pragma: no cover

    # from App
    @m.input()
    def allocate_code(self, length, wordlist):
        pass

    @m.input()
    def input_code(self):
        pass

    def set_code(self, code):
        validate_code(code)  # can raise KeyFormatError
        self._set_code(code)

    @m.input()
    def _set_code(self, code):
        pass

    # from Allocator
    @m.input()
    def allocated(self, nameplate, code):
        pass

    # from Input
    @m.input()
    def got_nameplate(self, nameplate):
        pass

    @m.input()
    def finished_input(self, code):
        pass

    @m.output()
    def do_set_code(self, code):
        nameplate = code.split("-", 2)[0]
        self._N.set_nameplate(nameplate)
        self._B.got_code(code)
        self._K.got_code(code)

    @m.output()
    def do_start_input(self):
        return self._I.start()

    @m.output()
    def do_middle_input(self, nameplate):
        self._N.set_nameplate(nameplate)

    @m.output()
    def do_finish_input(self, code):
        self._B.got_code(code)
        self._K.got_code(code)

    @m.output()
    def do_start_allocate(self, length, wordlist):
        self._A.allocate(length, wordlist)

    @m.output()
    def do_finish_allocate(self, nameplate, code):
        assert code.startswith(nameplate + "-"), (nameplate, code)
        self._N.set_nameplate(nameplate)
        self._B.got_code(code)
        self._K.got_code(code)

    S0_idle.upon(_set_code, enter=S4_known, outputs=[do_set_code])
    S0_idle.upon(
        input_code,
        enter=S1_inputting_nameplate,
        outputs=[do_start_input],
        collector=first)
    S1_inputting_nameplate.upon(
        got_nameplate, enter=S2_inputting_words, outputs=[do_middle_input])
    S2_inputting_words.upon(
        finished_input, enter=S4_known, outputs=[do_finish_input])
    S0_idle.upon(
        allocate_code, enter=S3_allocating, outputs=[do_start_allocate])
    S3_allocating.upon(allocated, enter=S4_known, outputs=[do_finish_allocate])
